# -*- coding: utf-8 -*-
"""

"""

from itertools import combinations
import numpy as npy
import networkx as nx
from genmechanics import geometry, tools
from scipy import linalg
from scipy.optimize import fsolve

from dessia_common import DessiaObject

import volmdlr as vm
import volmdlr.edges as edges

import webbrowser
import os
from jinja2 import Environment, PackageLoader, select_autoescape

class ModelError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'Model Error: ' + self.message

class Part(DessiaObject):
    _eq_is_data_eq=False
    
    
    def __init__(self, name='', interest_points=None):
        if interest_points is None:
            self.interest_points = []
        else:
            self.interest_points = interest_points
            
        self.name = name
    
        
    @classmethod
    def wireframe_lines(cls, points):
        if len(points) == 2:
            return [edges.LineSegment3D(points[0], points[1])]
        else:
            lines = []
            full_graph = nx.Graph()
            full_graph.add_nodes_from(points)
            points.append(vm.Point3D.mean_point(points))
            for point1, point2 in combinations(points, 2):
                full_graph.add_edge(point1, point2, weight=point1.point_distance(point2))
            
            wireframe_graph = nx.minimum_spanning_tree(full_graph)
            for point1, point2 in wireframe_graph.edges():
                lines.append(vm.LineSegment3D(point1, point2))
            
            return lines

class Mechanism:
    _eq_is_data_eq=False
    def __init__(self,
                 linkages,
                 ground,
                 imposed_speeds,
                 known_static_loads,
                 unknown_static_loads,
                 name=''):

        self.linkages = linkages
        self.ground = ground
        self.name = name
        self.imposed_speeds = imposed_speeds
        self.known_static_loads = known_static_loads
        self.unknown_static_loads = unknown_static_loads

        self._utd_kinematic_results = False
        self._utd_static_results = False

        self._holonomic_paths = {}


    def _get_parts(self):
        parts = []
        for linkage in self.linkages:
            for part in [linkage.part1,linkage.part2]:
                if not part in parts:
                    if part != self.ground:
                        parts.append(part)
        return parts

    parts = property(_get_parts)

    def _get_graph(self):
        G = nx.Graph()
        G.add_nodes_from(self.parts)
        for linkage in self.linkages:
            G.add_node(linkage)
            G.add_edge(linkage,linkage.part1)
            G.add_edge(linkage,linkage.part2)
        return G

    graph = property(_get_graph)

    def _get_holonomic_graph(self):
        G = nx.Graph()
        G.add_nodes_from(self.parts)
        for linkage in self.linkages:
            if linkage.holonomic:
                G.add_node(linkage)
                G.add_edge(linkage,linkage.part1)
                G.add_edge(linkage,linkage.part2)
        return G

    holonomic_graph = property(_get_holonomic_graph)

    def _settings_path(self, part1, part2):
        if (part1, part2) in self._settings_paths:
            return self._settings_paths[part1, part2]
        elif (part2, part1) in self._settings_paths:
            path = [(p2, linkage, not linkage_side, p1)\
                    for (p1, linkage, linkage_side, p2)\
                        in self._settings_paths[part2, part1][::-1]]
            self._settings_paths[part1, part2] = path
            return self._settings_paths[part1, part2]
        else:
            path = []
            raw_path = list(nx.shortest_path(self.settings_graph, part1, part2))
#            print('rp', raw_path)
            for part1, linkage, part2 in zip(raw_path[:-2:2], raw_path[1::2], raw_path[2::2]+[part2]):
                path.append((part1, linkage, linkage.part1==part1, part2))

            self._settings_paths[part1, part2] = path
            return path
        
    def settings_graph(self):
        graph = self.holonomic_graph.copy()
        deleted_linkages = []
        for cycle in nx.cycle_basis(graph):
            ground_distance = [(l, len(nx.shortest_path(graph, l, self.ground)))\
                               for l in cycle\
                               if l in self.linkages\
                                   and not l in deleted_linkages\
                                   and not l.positions_require_kinematic_parameters
                               ]
#            print(ground_distance)
            linkage_to_delete = max(ground_distance, key=lambda x:x[1])[0]
#            print(linkage_to_delete)
            deleted_linkages.append(linkage_to_delete)
            graph.remove_node(linkage_to_delete)
        
        self.linkages_kinematic_setting = [l for l in self.linkages if l not in deleted_linkages]
        self.settings_graph = graph

    def part_linkages(self):
        part_linkages = {}
        for linkage in self.linkages:
            for part in (linkage.part1, linkage.part2):
                if not part in part_linkages:
                    part_linkages[part] = [linkage]
                else:
                    part_linkages[part].append(linkage)
        return part_linkages


    def plot_graph(self):
        s="""<html>
        <head>
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.20.0/vis.min.js"></script>
        <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.20.0/vis.min.css" rel="stylesheet" type="text/css" />

        <style type="text/css">
            #mynetwork {
                border: 1px solid lightgray;
            }
        </style>
    </head>
    <body>
    <div id="mynetwork"></div>

    <script type="text/javascript">
    var nodes = new vis.DataSet([\n"""
        index={}
        for ipart,part in enumerate(self.parts+[self.ground]):
            index[part]=ipart
            s+="{{id: {}, label: '{}'}},\n".format(ipart,part.name)
#        s+=']);\n'
        n=len(self.parts)+1
#        index[self.ground]=n
#        n+=1
        for il,linkage in enumerate(self.linkages):
            index[linkage]=n+il
            s+="{{id: {}, label: '{}'}},\n".format(n+il,linkage.name)
        s+=']);\n'

        s+="var edges = new vis.DataSet(["
        for linkage in self.linkages:
            s+='{{from: {}, to: {}}},\n'.format(index[linkage],index[linkage.part1])
            s+='{{from: {}, to: {}}},\n'.format(index[linkage],index[linkage.part2])
        s+=']);'

        s+="""
    // create a network
    var container = document.getElementById('mynetwork');

    // provide the data in the vis format
    var data = {
        nodes: nodes,
        edges: edges
    };
    var options = {};

    // initialize your network!
    var network = new vis.Network(container, data, options);
</script>
</body>
</html>"""

        with open('gm_graph_viz.html','w') as file:
            file.write(s)

        webbrowser.open('file://' + os.path.realpath('gm_graph_viz.html'))

    def DrawPowerGraph(self,return_graph=False):
        """
        Draw graph with tulip
        """
        import matplotlib.pyplot as plt

        # Creating tulip graph
        G=nx.Graph()
#        nodes={}
#        edges={}
        widths=[]
        labels={}
        edges=[]
        for part in self.parts:
            G.add_node(part)

            labels[part]=part.name
#        nodes[self.ground]=G.addNode()
        for linkage in self.linkages:
            G.add_node(linkage)
            labels[linkage]=linkage.name
#            for part in [linkage.part1,linkage.part2]:
            G.add_edge(linkage,linkage.part1)
            widths.append(abs(self.TransmittedLinkagePower(linkage,0)))
            edges.append((linkage,linkage.part1))
            G.add_edge(linkage,linkage.part2)
            widths.append(abs(self.TransmittedLinkagePower(linkage,1)))
            edges.append((linkage,linkage.part2))

        for load in self.unknown_static_loads+self.known_static_loads:
            G.add_node(load)
            G.add_edge(load,load.part)
            widths.append(abs(self.LoadPower(load)))
            labels[load]=load.name
            edges.append((load,load.part))
        max_widths=max(widths)
#        print(widths)
        widths=[6*w/max_widths for w in widths]
#        edges[linkage,part]=e
        if not return_graph:
            plt.figure()
            pos=nx.spring_layout(G)
            nx.draw_networkx_nodes(G,pos,nodelist=self.linkages,node_color='grey')
            nx.draw_networkx_nodes(G,pos,nodelist=self.parts)
            nx.draw_networkx_nodes(G,pos,nodelist=self.unknown_static_loads,node_color='red')
            nx.draw_networkx_nodes(G,pos,nodelist=self.known_static_loads,node_color='green')
            nx.draw_networkx_nodes(G,pos,nodelist=self.parts,node_color='cyan')
            nx.draw_networkx_labels(G,pos,labels)
            nx.draw_networkx_edges(G,pos,edges,width=widths,edge_color='blue')
            nx.draw_networkx_edges(G,pos)
        else:
            return G
#        nx.draw_networkx_labels(G,pos,labels)

    def ChangeImposedSpeeds(self, imposed_speeds):
        self.imposed_speeds=imposed_speeds
        self._utd_kinematic_results=False
        self._utd_static_results=False# Change of speeds might change resistant forces: update needed

    def ChangeLoads(self,known_static_loads,unknown_static_loads):
        self.known_static_loads=known_static_loads
        self.unknown_static_loads=unknown_static_loads
        self._utd_static_results=False


    def Speeds(self,position,part_ref,part):
        """
        Speeds from point belonging to part with part_ref as reference
        """
        q=self.kinematic_vector#force kdof computation
        path=nx.shortest_path(self.holonomic_graph,part_ref,part)
        V=npy.zeros((3,self.n_kdof))
        W=npy.zeros((3,self.n_kdof))
        for il,linkage2 in enumerate(path):
            if not linkage2.__class__.__name__=='Part':
                try:
                    if path[il+1]==linkage2.part2:
                        side=1
                    else:
                        side=-1
                except IndexError:
                    # linkage is last element of list
                    if path[il-1]==linkage2.part1:
                        side=1
                    else:
                        side=-1
                # It's really a linkage
                P=geometry.Euler2TransferMatrix(*linkage2.euler_angles)
                u=linkage2.position
                uprime=u-position
                L=geometry.CrossProductMatrix(uprime)
                Ve=npy.dot(L,npy.dot(P,linkage2.kinematic_matrix[:3,:]))+npy.dot(P,linkage2.kinematic_matrix[3:,:])
                We=npy.dot(P,linkage2.kinematic_matrix[:3,:])
                for indof,ndof in enumerate(self.kdof[linkage2]):
                    V[:,ndof]+=side*Ve[:,indof]
                    W[:,ndof]+=side*We[:,indof]

        return npy.hstack((npy.dot(W,q).flatten(),npy.dot(V,q).flatten()))


    def LocalLinkageSpeeds(self,linkage):
        """
        :returns: relative speed of linkage in local linkage coordinate system
        """

        try:
            # holonomic linkage
            r=self.kinematic_results[linkage]# results
            vr=npy.array([r[i] for i in range(linkage.n_kinematic_unknowns)])#vector of results
            return npy.dot(linkage.kinematic_matrix,vr).flatten()
        except:
            s=self.Speed(linkage.position,self.ground,linkage.part1)
            v=npy.dot(linkage.P,s[:3])
            w=npy.dot(linkage.P,s[3:])
#            return npy.dot(linkage.P,s)
            return npy.hstack((w,v)).flatten()

    def LocalLinkageForces(self,linkage,num_part):
        """
        :returns :Local forces of linkage on part given by its number (0 for part1, 1 for part2)
        """
        r=self.static_results[linkage]# results
        vr=npy.array([r[i] for i in range(linkage.n_static_unknowns)])#vector of results
        if num_part==0:
            return npy.dot(linkage.static_matrix1,vr)
        else:
            return npy.dot(linkage.static_matrix2,vr)

    def GlobalLinkageForces(self,linkage,num_part):
        P=geometry.Euler2TransferMatrix(*linkage.euler_angles)
        lf=self.LocalLinkageForces(linkage,num_part)
        F=npy.zeros(6)
        F[:3]=npy.dot(P,lf[:3])
        F[3:]=npy.dot(P,lf[3:])
        return F

    def LocalLoadForces(self,load):
        try:
            F=npy.zeros(6)
            F[:3]=load.forces
            F[3:]=load.torques
            return F
        except AttributeError:
            r=self.static_results[load]# results
            vr=npy.array([r[i] for i in range(load.n_static_unknowns)])#vector of results
            return npy.dot(load.static_matrix,vr)


    def GlobalLoadForces(self,load):
        f=self.LocalLoadForces(load)
#        print(f)
        F=npy.zeros(6)
        F[:3]=npy.dot(load.P,f[:3]).flatten()
#        print(npy.dot(load.P,f[3:]).flatten())
        F[3:]=npy.dot(load.P,f[3:]).flatten()
        return F


    def LinkagePowerLosses(self,linkage):
        if linkage.holonomic:
            tf=self.LocalLinkageForces(linkage,0)
            ft=npy.hstack((tf[3:],tf[:3]))
            return npy.dot(self.LocalLinkageSpeeds(linkage),ft)
        else:
            d=self.GlobalLinkageForces(linkage,0)+self.GlobalLinkageForces(linkage,1)
            df=d[:3]
            dt=d[3:]
            s=-self.Speeds(linkage.position,self.ground,linkage.part1)
            w=s[:3]
            v=s[3:]
#            print(df,dt,w,v)
            return npy.dot(df,v)+npy.dot(dt,w)


    def LoadPower(self,load):
        s=self.Speeds(load.position,self.ground,load.part)
        f=self.GlobalLoadForces(load)
        P=npy.dot(s[3:],f[:3])
        P+=npy.dot(f[3:],s[:3])
        return P

    def TransmittedLinkagePower(self,linkage,num_part):
        if num_part==1:
            s=self.Speeds(linkage.position,self.ground,linkage.part2)
        else:
            s=self.Speeds(linkage.position,self.ground,linkage.part1)
        f=self.GlobalLinkageForces(linkage,num_part)
        P=npy.dot(s[3:],f[:3])
        P+=npy.dot(f[3:],s[:3])
        return P


    def _get_static_results(self):
        if not self._utd_static_results:
            self._static_results=self._StaticAnalysis()
            self._utd_static_results=True
        return self._static_results

    static_results=property(_get_static_results)

    def _get_kinematic_results(self):
        if not self._utd_kinematic_results:
            self._kinematic_results,self._kinematic_vector=self._KinematicAnalysis()
            self._utd_kinematic_results=True
        return self._kinematic_results

    kinematic_results=property(_get_kinematic_results)

    def _get_kinematic_vector(self):
        if not self._utd_kinematic_results:
            self._kinematic_results,self._kinematic_vector=self._KinematicAnalysis()
            self._utd_kinematic_results=True
        return self._kinematic_vector

    kinematic_vector=property(_get_kinematic_vector)


    def _StaticAnalysis(self):
        # Static parametrisation
        self.sdof={}
        self.n_sdof=0
        kinematic_analysis_required=False
        for linkage in self.linkages:
            self.sdof[linkage]=list(range(self.n_sdof,self.n_sdof+linkage.n_static_unknowns))
            self.n_sdof+=linkage.n_static_unknowns
            if linkage.static_require_kinematic:
                kinematic_analysis_required=True


        uloads_parts={}
        for load in self.unknown_static_loads:
            load_unknowns=load.static_matrix.shape[1]
            self.sdof[load]=list(range(self.n_sdof,self.n_sdof+load_unknowns))
            self.n_sdof+=load_unknowns
            if load.static_require_kinematic:
                kinematic_analysis_required=True

            # Loading sorting by part
            try:
                uloads_parts[load.part].append(load)
            except:
                uloads_parts[load.part]=[load]

        # Force kinematic computation if required
        if kinematic_analysis_required:
            self.kinematic_results
            

        # Knowns loads sorting by parts
        loads_parts={}
        for load in self.known_static_loads:
            try:
                loads_parts[load.part].append(load)
            except:
                loads_parts[load.part]=[load]



        lparts=len(self.parts)
        M=npy.zeros((6*lparts,self.n_sdof))
        K=npy.zeros((6*lparts,self.n_sdof))
        F=npy.zeros(6*lparts)
        q=npy.zeros(self.n_sdof)
        nonlinear_eq={}

        # Occurence matrix assembly
        for ip,part in enumerate(self.parts):
            # linkage contribution to the LHS
            for linkage in self.graph[part].keys():
                P=geometry.Euler2TransferMatrix(*linkage.euler_angles)
                u=linkage.position
                uprime=u
                L=geometry.CrossProductMatrix(uprime)
                if part==linkage.part1:
                    static_matrix=linkage.static_matrix1
                else:
                    static_matrix=linkage.static_matrix2

                Me1=npy.abs(npy.dot(P,static_matrix[:3,:]))>1e-10
                Me2=npy.abs(npy.dot(L,npy.dot(P,static_matrix[:3,:]))+npy.dot(P,static_matrix[3:,:]))>1e-10
                Me=npy.vstack([Me1,Me2])
                for indof,ndof in enumerate(self.sdof[linkage]):
                    M[ip*6:(ip+1)*6,ndof]+=Me[:,indof]


                Ke1=npy.dot(P,static_matrix[:3,:])
                Ke2=npy.dot(L,npy.dot(P,static_matrix[:3,:]))+npy.dot(P,static_matrix[3:,:])
                Ke=npy.vstack([Ke1,Ke2])
                for indof,ndof in enumerate(self.sdof[linkage]):
                    K[ip*6:(ip+1)*6,ndof]+=Ke[:,indof]


            # Unknowns loads contribution to the LHS
            try:
                uloads=uloads_parts[part]
            except:
                uloads=[]
            for load in uloads:
                P=geometry.Euler2TransferMatrix(*load.euler_angles)
                u=load.position
                uprime=u
                L=geometry.CrossProductMatrix(uprime)
                Me1=npy.abs(npy.dot(P,load.static_matrix[:3,:]))>1e-10
                Me2=npy.abs(npy.dot(L,npy.dot(P,load.static_matrix[:3,:]))+npy.dot(P,load.static_matrix[3:,:]))>1e-10
                Me=npy.vstack([Me1,Me2])
                for indof,ndof in enumerate(self.sdof[load]):
                    M[ip*6:(ip+1)*6,ndof]+=Me[:,indof]

                Ke1=npy.dot(P,load.static_matrix[:3,:])
                Ke2=npy.dot(L,npy.dot(P,load.static_matrix[:3,:]))+npy.dot(P,load.static_matrix[3:,:])
                Ke=npy.vstack([Ke1,Ke2])
                for indof,ndof in enumerate(self.sdof[load]):
                    K[ip*6:(ip+1)*6,ndof]=Ke[:,indof]     # minus because of sum is in LHS


            # knowns loads contribution to the RHS
            try:
                loads=loads_parts[part]
            except:
                loads=[]
            for load in loads:
                P=geometry.Euler2TransferMatrix(*load.euler_angles)
                u=load.position
                uprime=u
                L=geometry.CrossProductMatrix(uprime)
                F1=npy.dot(P,load.forces)
                F2=npy.dot(L,npy.dot(P,load.forces))+npy.dot(P,load.torques)
                Fe=npy.hstack([F1.reshape(3),F2.reshape(3)])
                F[ip*6:(ip+1)*6]-=Fe # minus because of sum is in LHS

        neq=6*lparts
        neq_linear=neq
        indices_r=list(range(neq))

        # behavior equations of linkages
        for linkage in self.linkages:
            neq_linkage=linkage.static_behavior_occurence_matrix.shape[0]
            if neq_linkage>0:
                # Adding to occurence matrix
                Me=npy.zeros((neq_linkage,self.n_sdof))
                for indof,ndof in enumerate(self.sdof[linkage]):
                    Me[:,ndof]+=linkage.static_behavior_occurence_matrix[:,indof]
                M=npy.vstack([M,Me])

                # Adding linear equations to System matrix K
                neq_linear_linkage=linkage.static_behavior_linear_eq.shape[0]
                if neq_linear_linkage>0:
                    Ke=npy.zeros((neq_linear_linkage,self.n_sdof))
                    for indof,ndof in enumerate(self.sdof[linkage]):
                        Ke[neq_linear:neq_linear+6,ndof]+=linkage.static_behavior_linear_eq[:,indof]
                    K=npy.vstack([K,Ke])
                    indices_r.extend(range(neq_linear,neq_linear+neq_linear_linkage))

                # Collecting non linear equations
#                print(v)
                if linkage.holonomic:
                    if linkage.static_require_kinematic:
                        # Relatives speed in this case
                        ls=self.LocalLinkageSpeeds(linkage)
                        w=ls[:3]
                        v=ls[3:]
                    else:
                        w,v=0,0
                    for i,fct in zip(linkage.static_behavior_nonlinear_eq_indices,linkage.static_behavior_nonlinear_eq):
                        nonlinear_eq[neq+i]=lambda x,v=v,w=w,fct=fct:fct(x,w,v)
                else:
                    if linkage.static_require_kinematic:
                        # Absolute speed in this case in local coordinate system
                        s=self.Speeds(linkage.position,self.ground,linkage.part1)
                        w=npy.dot(linkage.P.T,s[:3])
                        v=npy.dot(linkage.P.T,s[3:])
                    else:
                        w,v=0,0

                    for i,fct in zip(linkage.static_behavior_nonlinear_eq_indices,linkage.static_behavior_nonlinear_eq):
                        nonlinear_eq[neq+i]=lambda x,v=v,w=w,fct=fct:fct(x,w,v)

                # Updating counters
                neq+=neq_linkage
                neq_linear+=neq_linear_linkage



        # behavior equations of unknowns loads
        for load in self.unknown_static_loads:
            neq_load=load.static_behavior_occurence_matrix.shape[0]
            if neq_load>0:
                # Adding to occurence matrix
                Me=npy.zeros((neq_load,self.n_sdof))
                for indof,ndof in enumerate(self.sdof[load]):
                    Me[:,ndof]+=load.static_behavior_occurence_matrix[:,indof]
                M=npy.vstack([M,Me])

                # Adding linear equations to System matrix K
                neq_linear_load=load.static_behavior_linear_eq.shape[0]
                if neq_linear_load>0:
                    Ke=npy.zeros((neq_linear_load,self.n_sdof))
                    for indof,ndof in enumerate(self.sdof[load]):
                        Ke[neq_linear:neq_linear+6,ndof]+=load.static_behavior_linear_eq[:,indof]
                    K=npy.vstack([K,Ke])
                    indices_r.extend(range(neq_linear,neq_linear+neq_linear_load))

                # Collecting non linear equations

                if load.static_require_kinematic:
                    # Absolute speed in this case in local coordinate system
                    s=self.Speeds(load.position,self.ground,load.part)
                    w=npy.dot(load.P.T,s[:3])
                    v=npy.dot(load.P.T,s[3:])
                else:
                    w,v=0,0
                for i,fct in zip(load.static_behavior_nonlinear_eq_indices,load.static_behavior_nonlinear_eq):
                    nonlinear_eq[neq+i]=lambda x,v=v,w=w,fct=fct:fct(x,w,v)

                # Updating counters
                neq+=neq_load
                neq_linear+=neq_linear_load


        solvable,solvable_var,resolution_order=tools.EquationsSystemAnalysis(M,None)
#        print(resolution_order)
        if not solvable:
            raise ModelError('Overconstrained system')
        for eqs,variables in resolution_order:
#            print(eqs,variables)
            linear=True
            linear_eqs=[]
            for eq in eqs:
                try:
                    nonlinear_eq[eq]
                    linear=False
                except KeyError:
                    linear_eqs.append(eq)
            if linear:
                eqs_r=npy.array([indices_r[eq] for eq in eqs])
                other_vars=npy.array([i for i in range(self.n_sdof) if i not in variables])
                Kr=K[eqs_r[:,None],npy.array(variables)]
                Fr=F[eqs_r]-npy.dot(K[eqs_r[:,None],other_vars],q[other_vars])

                q[variables]=linalg.solve(Kr,Fr)
            else:
                nl_eqs=[]
                other_vars=npy.array([i for i in range(self.n_sdof) if i not in variables])
                for eq in eqs:
                    try:
                        f1=nonlinear_eq[eq]
                        vars_func=[i for i in range(self.n_sdof) if M[eq,i]]
                        def f2(x,f1=f1,vars_func=vars_func,variables=variables,q=q):
                            x2=[]
                            for variable in vars_func:
                                try:
                                    x2.append(x[variables.index(variable)])
                                except ValueError:
                                    x2.append(q[variable])
                            return f1(x2)
                        nl_eqs.append(f2)
                    except KeyError:
                        # lambdification of linear equations
                        f2=lambda x,indices_r=indices_r,K=K,F=F,eq=eq,q=q,other_vars=other_vars:npy.dot(K[indices_r[eq],variables],x)-F[indices_r[eq]]+npy.dot(K[indices_r[eq],other_vars],q[other_vars])
                        nl_eqs.append(f2)
                f=lambda x:[fi(x) for fi in nl_eqs]
                xs=fsolve(f,npy.zeros(len(variables)),full_output=0)
                if npy.sum(npy.abs(f(xs)))>1e-4:
                    raise ModelError('No convergence of nonlinear phenomena solving'+str(npy.sum(npy.abs(f(xs)))))
                q[variables]=xs


        results={}
        for link,dofs in self.sdof.items():
            rlink={}
            for idof,dof in enumerate(dofs):
                rlink[idof]=q[dof]
            results[link]=rlink

        return results

    def _KinematicAnalysis(self):
        # Kinematic setting
        self.n_kdof = 0
        self.kdof = {}
        loops = nx.cycle_basis(self.holonomic_graph)
        ll = len(loops)
        ieq = 6*ll
        neq = ieq+len(self.imposed_speeds)
        nhl = []
        for linkage in self.linkages:
            try:
                # Holonomic linkage
                ndofl=linkage.kinematic_matrix.shape[1]
                for dofl in range(ndofl):
                    try:
                        self.kdof[linkage].append(self.n_kdof+dofl)
                    except KeyError:
                        self.kdof[linkage]=[self.n_kdof+dofl]
                self.n_kdof+=ndofl
            except AttributeError:
                # Non holonomic linkage
                nhl.append(linkage)
                neq+=len(linkage.kinematic_directions)

        K = npy.zeros((neq,self.n_kdof))
        F = npy.zeros((neq,1))
        q = npy.zeros((self.n_kdof,1))
        M = npy.zeros((neq,self.n_kdof))

        for il,loop in enumerate(loops):
            for ilk,linkage in enumerate(loop):
                if not linkage.__class__.__name__ == 'Part':
                    try:
                        if loop[ilk+1] == linkage.part2:
                            side = 1
                        else:
                            side =- 1
                    except IndexError:
                        # linkage is last element of list
                        if loop[ilk-1] == linkage.part1:
                            side = 1
                        else:
                            side =- 1
                    # Linkage
                    P=geometry.Euler2TransferMatrix(*linkage.euler_angles)
                    u=linkage.position
                    uprime=u
                    L=geometry.CrossProductMatrix(uprime)
                    Me1=npy.abs(npy.dot(P,linkage.kinematic_matrix[:3,:]))>1e-10
                    Me2=npy.abs(npy.dot(L,npy.dot(P,linkage.kinematic_matrix[:3,:]))+npy.dot(P,linkage.kinematic_matrix[3:,:]))>1e-10
                    Me=npy.vstack([Me1,Me2])
                    for indof,ndof in enumerate(self.kdof[linkage]):
                        M[6*il:6*il+6,ndof] += Me[:,indof]


                    Ke1=npy.dot(P,linkage.kinematic_matrix[:3,:])
                    Ke2=npy.dot(L,npy.dot(P,linkage.kinematic_matrix[:3,:]))+npy.dot(P,linkage.kinematic_matrix[3:,:])
                    Ke=side*npy.vstack([Ke1,Ke2])
                    for indof,ndof in enumerate(self.kdof[linkage]):
                        K[6*il:6*il+6,ndof]+=Ke[:,indof]

        # Non holonomic equations
        for linkage in nhl:
            # Speed computation
            try:
                path = nx.shortest_path(self.holonomic_graph,
                                        linkage.part1,
                                        linkage.part2)
            except nx.NetworkXNoPath:
                raise ModelError('No path between {} and {} for linkage {} of type {}'.format(linkage.part1.name, linkage.part2.name, linkage.name, linkage.__class__.__name__))

            V = npy.zeros((3, self.n_kdof))
            for il, linkage2 in enumerate(path):
                if not linkage2.__class__.__name__ == 'Part':
                    try:
                        if path[il+1] == linkage2.part2:
                            side = 1
                        else:
                            side = -1
                    except IndexError:
                        # linkage is last element of list
                        if path[il-1] == linkage2.part1:
                            side = 1
                        else:
                            side = -1
                    # It's really a linkage
                    P = geometry.Euler2TransferMatrix(*linkage2.euler_angles)
                    u = linkage2.position
                    uprime = u-linkage.position
                    L = geometry.CrossProductMatrix(uprime)
                    Ve = npy.dot(L,npy.dot(P,linkage2.kinematic_matrix[:3,:]))+npy.dot(P,linkage2.kinematic_matrix[3:,:])
                    for indof, ndof in enumerate(self.kdof[linkage2]):
                        V[:,ndof] += side*Ve[:,indof]


            P = geometry.Euler2TransferMatrix(*linkage.euler_angles)
            for direction in linkage.kinematic_directions:
                K[ieq,:] = npy.dot(npy.dot(P, direction), V)
                ieq+=1

        # Imposed speeds equations
        for linkage,index,speed in self.imposed_speeds:
            K[ieq, self.kdof[linkage][index]] = 1
            F[ieq, 0] = speed
            ieq += 1

        # deducing M from K for last lines
        M[6*ll:, :] = npy.abs(K[6*ll:, :]) > 1e-10

        solvable, solvable_var, resolution_order = tools.EquationsSystemAnalysis(M, None)

        if solvable:
            for eqs,variables in resolution_order:
                eqs = npy.array(eqs)
                other_vars = npy.array([i for i in range(self.n_kdof) if i not in variables])
                Kr = K[eqs[:,None],npy.array(variables)]
                Fr = F[eqs,:]-npy.dot(K[eqs[:,None], other_vars], q[other_vars, :])
                q[variables, :] = linalg.solve(Kr,Fr)
            results = {}
            for link, dofs in self.kdof.items():
                rlink = {}
                for idof, dof in enumerate(dofs):
                    rlink[idof] = q[dof, 0]
                results[link] = rlink
            return results, q
        else:
            raise ModelError

    def GlobalSankey(self):
        from matplotlib.sankey import Sankey
        flows=[]
        orientations=[]
        labels=[]


        for load in self.known_static_loads+self.unknown_static_loads:
            pl=self.LoadPower(load)
#            print(pl)
            flows.append(pl)
            labels.append(load.name)
            if (load.__class__.__name__=='SimpleUnknownLoad')|(load.__class__.__name__=='KnownLoad'):
                orientations.append(0)
            else:
                orientations.append(-1)

        for linkage in self.linkages:
            pl=self.LinkagePowerLosses(linkage)
#            print(pl)
            if pl!=0.:
                flows.append(-pl)
                orientations.append(-1)
                labels.append(linkage.name)
#            pl.append(0.5*l)
        l = max([abs(f) for f in flows])
#        pl=[0.1*l]*len(flows)

        sankey = Sankey(unit='W',scale=1/l)

        sankey.add(flows=flows,
           orientations=orientations,labels=labels,)

        sankey.finish()

    def VMPlot(self, u=1):
        points=[]
        for linkage in self.linkages:
            points.append(vm.Point3D(linkage.position))
            points.append(vm.Line3D(vm.Point3D(linkage.position),vm.Point3D(linkage.position+u*linkage.P[:,0])))
        mdl=vm.VolumeModel(points)
        mdl.MPLPlot()

    def SceneCaracteristics(self):
        min_vect=self.linkages[0].position.copy()
        max_vect=self.linkages[0].position.copy()
        center=self.linkages[0].position.copy()
        n=1
        for linkage in self.linkages[1:]+self.known_static_loads+self.unknown_static_loads:
#            print(linkage)
            for i,(xmin,xmax,xi) in enumerate(zip(min_vect,max_vect,linkage.position)):
#                print(i,xmin,xmax,xi)
                if xi<xmin:
                    min_vect[i]=xi
#                    print('min',min_vect)
                if xi>xmax:
                    max_vect[i]=xi
#                    print('max',max_vect)
#            print(linkage,linkage.position)
            center+=linkage.position
            n+=1

        center=center/n
#        print(min_vect,max_vect)
        max_length=linalg.norm(min_vect-max_vect)
#        print(center,max_length)
        return center,max_length


    def BabylonScript(self,forces=True):

        env = Environment(loader=PackageLoader('genmechanics', 'templates'),
                          autoescape=select_autoescape(['html', 'xml']))

        template = env.get_template('babylon.html')

        center,length=self.SceneCaracteristics()

        if forces:
            max_force=0.
            max_torque=0
            for linkage in self.linkages:
                f=self.GlobalLinkageForces(linkage,0)[0:3]
                t=self.GlobalLinkageForces(linkage,0)[3:]
                max_force=max(max_force,max(f))
                max_torque=max(max_torque,max(t))

            for load in self.known_static_loads+self.unknown_static_loads:
                f=self.GlobalLoadForces(load)[0:3]
                t=self.GlobalLoadForces(load)[3:]
                max_force=max(max_force,max(f))
                max_torque=max(max_torque,max(t))

#            print(max_force,max_torque)


        linkages_strings=[]
        for linkage in self.linkages:
            try:
                if forces:
#                    print([f/max_force/length/4 for f in self.GlobalLoadForces(load)[0:3]])
                    linkages_strings.append(linkage.Babylon(length,
                                                            [f/max_force*length/4 for f in self.GlobalLinkageForces(linkage,0)[0:3]],
                                                            [f/max_torque*length/4 for f in self.GlobalLinkageForces(linkage,0)[3:]]))
                else:
                    linkages_strings.append(linkage.Babylon(length,None,None))

            except AttributeError:
                print('error')
                pass




        return template.render(name=self.name,center=tuple(center),length=length,
                               linkages_strings=linkages_strings)

    def BabylonShow(self,page='gm_babylonjs',forces=True):
        page+='.html'
#        print(self.BabylonScript())
        with open(page,'w') as file:
            file.write(self.BabylonScript())

        webbrowser.open('file://' + os.path.realpath(page))

#    def BabylonJS(self):
#        center,length=self.SceneCaracteristics()
#        s="""
#        <!doctype html>
#<html>
#<head>
#   <meta charset="utf-8">
#   <title>Babylon - Basic scene</title>
#   <style>
#      html, body {
#         overflow: hidden;
#         width: 100%;
#         height: 100%;
#         margin: 0;
#         padding: 0;
#      }
#      #renderCanvas {
#         width: 100%;
#         height: 100%;
#         touch-action: none;
#      }
#   </style>
#   <script src="https://cdnjs.cloudflare.com/ajax/libs/babylonjs/2.5.0/babylon.js"></script>
#   <script src="https://cdnjs.cloudflare.com/ajax/libs/handjs/1.3.11/hand.js"></script>
#   <script src="https://cdnjs.cloudflare.com/ajax/libs/cannon.js/0.6.2/cannon.min.js"></script> <!-- optional physics engine -->
#</head>
#<body>
#   <canvas id="renderCanvas"></canvas>
#   <script type="text/javascript">
#      // Get the canvas element from our HTML below
#      var canvas = document.querySelector("#renderCanvas");
#      // Load the BABYLON 3D engine
#      var engine = new BABYLON.Engine(canvas, true);
#      // -------------------------------------------------------------
#      // Here begins a function that we will 'call' just after it's built
#      var createScene = function () {
#         // Now create a basic Babylon Scene object
#         var scene = new BABYLON.Scene(engine);
#         // This creates and positions a free camera
#         //var plane = BABYLON.Mesh.CreatePlane("plane", 10.0, scene);
#         """
#
#
##        s+='var camera = new BABYLON.FreeCamera("camera1", new BABYLON.Vector3({}, {}, {}), scene);'.format(*center)
#        s+='var camera = new BABYLON.ArcRotateCamera("ArcRotateCamera", 0., 0., {}, new BABYLON.Vector3({}, {}, {}), scene);\n'.format(4*length,*center)
##        s+='var camera = new BABYLON.ArcRotateCamera("ArcRotateCamera", 0., 0., 15, new BABYLON.Vector3(0,0,0), scene);'
#        s+='camera.panningSensibility={};\n'.format(30)
#        s+='camera.pinchPrecision={};\n'.format(20)
#        s+='camera.wheelPrecision={};\n'.format(20)
#        s+="""
#        camera.attachControl(canvas, false);
#         // This creates a light, aiming 0,1,0 - to the sky.
#         var light = new BABYLON.HemisphericLight("light1", new BABYLON.Vector3(0, 1, 0), scene);
#         // Dim the light a small amount
#         light.intensity = .5;
#         """
#
##         // Let's try our built-in 'sphere' shape. Params: name, subdivisions, size, scene
##         var sphere = BABYLON.Mesh.CreateSphere("sphere1", 16, 2, scene);
##         // Move the sphere upward 1/2 its height
##         sphere.position.y = 1;
##         // Let's try our built-in 'ground' shape. Params: name, width, depth, subdivisions, scene
##         var ground = BABYLON.Mesh.CreateGround("ground1", 6, 6, 2, scene);
#
#        for linkage in self.linkages:
#            s+='var sphere = BABYLON.Mesh.CreateSphere("sphere1", 15., {}, scene);\n'.format(length/30)
#            s+="sphere.position=new BABYLON.Vector3({},{},{});\n".format(*linkage.position)
##            s+="var lines = BABYLON.Mesh.CreateLines("lines", [    new BABYLON.Vector3(-10, 0, 0)"
#        s+="""
#         // Leave this function
#         return scene;
#      }; // End of createScene function
#      // -------------------------------------------------------------
#      // Now, call the createScene function that you just finished creating
#      var scene = createScene();
#      // Register a render loop to repeatedly render the scene
#      engine.runRenderLoop(function () {
#         scene.render();
#      });
#      // Watch for browser/canvas resize events
#      window.addEventListener("resize", function () {
#         engine.resize();
#      });
#   </script>
#</body>
#</html>
#
#        """
#        with open('gm_babylonjs.html','w') as file:
#            file.write(s)
#
#        webbrowser.open('file://' + os.path.realpath('gm_babylonjs.html'))




# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 20:13:02 2016

@author: steven
"""

import networkx as nx

def EquationsSystemAnalysis(Mo,vars_to_solve,overconstrain_stop=True):
    """
        Analyse a free equations system given by its ocurence matrix.
        :return: False (if system is unsolvable) if overconstrain_stop==True
        else, returns True, the solvable variables and the resolution order
    """
    
    if vars_to_solve==None:
        vars_to_solve=range(Mo.shape[1])
    
    
    neq,nvar=Mo.shape
    G=nx.Graph()
    Gp=nx.DiGraph()
    pos={}
    for i in range(nvar):
        G.add_node('v'+str(i),bipartite=0)
        Gp.add_node('v'+str(i),bipartite=0)
        pos['v'+str(i)]=[i,0]
    
    for i in range(neq):
        G.add_node('e'+str(i),bipartite=1)
        Gp.add_node('e'+str(i),bipartite=1)
        pos['e'+str(i)]=[i,1]
        for j in range(nvar):
            if Mo[i,j]==1:
                G.add_edge('e'+str(i),'v'+str(j))
                Gp.add_edge('e'+str(i),'v'+str(j))
            
        
    
#    plt.figure()
#    pos=nx.spring_layout(G)
#    nx.draw(G,pos)
#    nx.draw_networkx_labels(G,pos)
               
    for Gi in (G.subgraph(c).copy() for c in nx.connected_components(G)):
        M=nx.bipartite.maximum_matching(Gi)
    #    print('M',M,len(M))
        
        for n1,n2 in M.items():
    #        print(n1,n2)
            Gp.add_edge(n1,n2)    
    
#    pos=nx.spring_layout(Gp)
#    plt.figure()
#    nx.draw(Gp,pos)
#    nx.draw_networkx_labels(Gp,pos)
    
    sinks=[]
    sources=[]
    for node in Gp.nodes():
        if Gp.out_degree(node)==0:
            sinks.append(node)
        elif Gp.in_degree(node)==0:
            sources.append(node)
    
    G2=sources[:]
    for node in sources:
        for node2 in nx.descendants(Gp,node):
            if node2 not in G2:
                G2.append(node2)
    
    if overconstrain_stop:
        if G2!=[]:
#            print('G2 (sur-contraint): ',G2)
#            eG2=[int(elem[1:]) for elem in G2 if elem[0]=='e']
#            vG2=[int(elem[1:]) for elem in G2 if elem[0]=='v']
#            print(Mo[eG2,vG2])
            return (False,[],None)

    
    G3=sinks[:]
    for node in sinks:
        for node2 in nx.ancestors(Gp,node):
            if node2 not in G3:
                G3.append(node2)

#    print(sources)
#    print(sinks)
    
#    print('G3 (sous-contraint): ',G3)

    solvable_vars=[]
    for var in vars_to_solve: 
        if not 'v'+str(var) in G2+G3:
            solvable_vars.append(var)
            
        
    G1=G.copy()
    G1.remove_nodes_from(G2+G3)
#    print('G1: ',G1.nodes())

#    pos=nx.spring_layout(G1)
#    plt.figure()
#    nx.draw(G1,pos)
#    nx.draw_networkx_labels(G1,pos)
    
#    M1=nx.bipartite.maximum_matching(G1)
    G1p=nx.DiGraph()

    G1p.add_nodes_from(G1.nodes())
    for e in G1.edges():
        # equation vers variable
        if e[0][0]=='v':
            G1p.add_edge(e[0],e[1])        
        else:
            G1p.add_edge(e[1],e[0])   

    
    for G1i in (G1.subgraph(c).copy() for c in nx.connected_components(G1)):#nx.connected_component_subgraphs(G1):
        M1=nx.bipartite.maximum_matching(G1i)
    

        for n1,n2 in M1.items():
    #        print(n1,n2)
            if n1[0]=='e':
                G1p.add_edge(n1,n2)        
            else:
                G1p.add_edge(n2,n1)
            
        
    scc=list(nx.strongly_connected_components(G1p))
#    pos=nx.spring_layout(G1p)
#    plt.figure()
#    nx.draw(G1p,pos)
#    nx.draw_networkx_labels(G1p,pos)
    
    if scc!=[]:
        C=nx.condensation(G1p,scc)
#        print('scc: ',scc)
#        plt.figure()
#        pos=nx.spectral_layout(C)
#        nx.draw(C,pos)
#        nx.draw_networkx_labels(C,pos)
        # on cherche les indices des blocs contenant chacune des variables
#        print(list(C.nodes()))
        isc_vars=[]
        for isc,sc in enumerate(scc):
#            print(sc)
            for var in solvable_vars:
                if 'v'+str(var) in sc:
                    isc_vars.append(isc)
                    break
#                print(isc)
        ancetres_vars=isc_vars[:]
        
        for isc_var in isc_vars:                                
            for ancetre in nx.ancestors(C,isc_var):
                if ancetre not in ancetres_vars:
                    ancetres_vars.append(ancetre)
                
#        ancetres_var.add(isc_var)
#        print('ancetres: ',ancetres_var)
#        ordre_sc=nx.topological_sort_recursive(C,reverse=False)
#        print(ancetres_vars)
        ordre_sc=[sc for sc in nx.topological_sort(C) if sc in ancetres_vars]
#        print(ordre_sc)
        ordre_ev=[]
        for isc in ordre_sc:            
            evs=sorted(scc[isc])# liste d'équations et de variables triées pour être séparées

            levs=int(len(evs)/2)
#            print(evs,levs,levs/2)
            ordre_ev.append(([int(e[1:]) for e in evs[0:levs]],[int(v[1:]) for v in evs[levs:]]))
            
#        print(ordre_ev)        
        return (True,solvable_vars,ordre_ev)
        
    return (False,[],None)


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 11:44:25 2016

@author: steven
"""
import numpy as npy
import math
import scipy.linalg as linalg

def CrossProductMatrix(u):
    L=npy.array([[0,-u[2],u[1]],[u[2],0,-u[0]],[-u[1],u[0],0]])
    return L
    
def Euler2TransferMatrix(psi,theta,phi):
    """
    Give Transition Matrix from euler angles
    Angles in radians
    """

    cpsi=math.cos(psi)
    spsi=math.sin(psi)
    ctheta=math.cos(theta)
    stheta=math.sin(theta)
    cphi=math.cos(phi)
    sphi=math.sin(phi)
    P=npy.array([[cphi*cpsi-sphi*ctheta*spsi,-spsi*cphi-cpsi*ctheta*sphi,stheta*sphi],[cpsi*sphi+spsi*ctheta*cphi,-sphi*spsi+cphi*ctheta*cpsi,-stheta*cphi],[spsi*stheta,cpsi*stheta,ctheta]])
    return P
    

def TransferMatrix2Euler(R):
    if ((R[2,2]!=1)&(R[2,2]!=-1)):
#        R[2,2]=R[2,2]/abs(R[2,2])
        theta=math.acos(R[2,2])
        psi=math.atan2(R[2,0]/math.sin(theta),R[2,1]/math.sin(theta))
        phi=math.atan2(R[0,2]/math.sin(theta),-R[1,2]/math.sin(theta))
    else:
        phi=0
        if R[2,2]==1:
            theta=0
            psi=math.atan2(R[1,0],R[0,0])
        else:
            theta=math.pi
            psi=-math.atan2(R[1,0],R[0,0])
    return(npy.array([psi,theta,phi]))

    
def Direction2Euler(u,v=npy.random.random(3)):
#    u=npy.array([ux,uy,uz])
    u=u/linalg.norm(u)
    R=npy.zeros((3,3))
    R[:,0]=u
    v=v-npy.dot(u,v)*u
    v=v/linalg.norm(v)
    w=npy.cross(u,v)
    R[:,1]=v
    R[:,2]=w
    euler=TransferMatrix2Euler(R)
    return euler


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 13:14:34 2016

@author: Steven Masfaraud
"""

import numpy as npy
from math import cos,sin
import genmechanics.geometry as geometry
from dessia_common import DessiaObject
class Linkage:
    _eq_is_data_eq=False
    _non_serializable_attributes=['position','euler_angles',
                 'static_matrix1', 'static_matrix2',
                 'static_behavior_occurence_matrix',
                 'static_behavior_nonlinear_eq_indices',
                 'static_behavior_linear_eq',
                 'static_behavior_nonlinear_eq',
                 'static_require_kinematic','P','kinematic_matrix']
    def __init__(self,
                 part1, part2, position,
                 euler_angles,
                 static_matrix1, static_matrix2,
                 static_behavior_occurence_matrix,
                 static_behavior_nonlinear_eq_indices,
                 static_behavior_linear_eq,
                 static_behavior_nonlinear_eq,
                 static_require_kinematic,name=''):
        
       
        self.part1 = part1
        self.part2 = part2
        self.position = npy.array(position)
        self.euler_angles = euler_angles
        self.name = name        
        
        self.static_matrix1=static_matrix1
        self.static_matrix2=static_matrix2
        self.static_behavior_occurence_matrix=static_behavior_occurence_matrix
        
        self.static_behavior_nonlinear_eq_indices=static_behavior_nonlinear_eq_indices
        self.static_behavior_linear_eq=static_behavior_linear_eq
        self.static_behavior_nonlinear_eq=static_behavior_nonlinear_eq
        self.static_require_kinematic=static_require_kinematic
        
        self.n_static_unknowns=static_matrix1.shape[1]
                
        self.P=geometry.Euler2TransferMatrix(*self.euler_angles) 
        
    def Babylon(self,length,forces,torques):
        xa,za,ya=self.P[:,0]
#        theta=math.acos(z/self.width)
#        phi=math.atan(y/x)
        x,z,y=self.position
        s="""
        var sphere = BABYLON.Mesh.CreateSphere("{} center", 8, {}, scene);
        sphere.position=new BABYLON.Vector3({},{},{});    
        var lineX = BABYLON.Mesh.CreateDashedLines("{} axis", [
            new BABYLON.Vector3({}, {}, {}),
            new BABYLON.Vector3({}, {}, {})
        ],0.05,0.05,10, scene);
        """.format(self.name,length/30,x,y,z,self.name,x-0.5*xa,y-0.5*ya,z-0.5*za,x+0.5*xa,y+0.5*ya,z+0.5*za)
        
        if forces is not None:
#            print(forces)
#            for i,fi in enumerate(forces):
            s+="""var lineF = BABYLON.Mesh.CreateLines("{} axis", [
            new BABYLON.Vector3({}, {}, {}),
            new BABYLON.Vector3({}, {}, {})
        ], scene);
            lineF.color=new BABYLON.Color3(1,0,0);""".format('force {}'.format(self.name),x,y,z,x+forces[0],y+forces[1],z+forces[2])
                
        
        return s
        
        

class HolonomicLinkage(Linkage):
    holonomic = True
    
    def __init__(self,part1,part2,position,euler_angles,static_matrix1,static_matrix2,
                 static_behavior_occurence_matrix,static_behavior_nonlinear_eq_indices,
                 static_behavior_linear_eq,static_behavior_nonlinear_eq,
                 kinematic_matrix,static_require_kinematic=False,name=''):
        Linkage.__init__(self,part1,part2,position,euler_angles,static_matrix1,static_matrix2,
                         static_behavior_occurence_matrix,static_behavior_nonlinear_eq_indices,
                         static_behavior_linear_eq,static_behavior_nonlinear_eq,
                         static_require_kinematic,name)
        self.kinematic_matrix=kinematic_matrix
        self.n_kinematic_unknowns=kinematic_matrix.shape[1]
        
    
        
class NonHolonomicLinkage(Linkage):
    holonomic = False
    
    def __init__(self,
                 part1, part2, position, euler_angles,
                 static_matrix1, static_matrix2,
                 static_behavior_occurence_matrix, 
                 static_behavior_nonlinear_eq_indices,
                 static_behavior_linear_eq, static_behavior_nonlinear_eq,
                 kinematic_directions,
                 static_require_kinematic=False, name=''):
        
        Linkage.__init__(self,part1,part2,position,euler_angles,static_matrix1,static_matrix2,
                         static_behavior_occurence_matrix,static_behavior_nonlinear_eq_indices,
                         static_behavior_linear_eq,static_behavior_nonlinear_eq,
                         static_require_kinematic,name)
        self.kinematic_directions=kinematic_directions
        

        
class FrictionlessRevoluteLinkage(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,name='Frictionless Revolute Linkage'):
        static_matrix2=npy.array([[1,0,0,0,0],[0,1,0,0,0],[0,0,1,0,0],[0,0,0,0,0],[0,0,0,1,0],[0,0,0,0,1]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([])
        static_behavior_nonlinear_eq_indices=[]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[]
        kinematic_matrix=npy.array([[1],[0],[0],[0],[0],[0]])
        static_require_kinematic=False
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
                                  static_behavior_nonlinear_eq,kinematic_matrix,static_require_kinematic,name)
        # print(self.to_dict())
        # print('EZA')

class RevoluteLinkage(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,Ca,Cr,Cw,name='Revolute Linkage'):
        self.Ca=Ca
        self.Cr=Cr
        self.Cw=Cw
        static_matrix2=npy.array([[1,0,0,0,0,0],[0,1,0,0,0,0],[0,0,1,0,0,0],[0,0,0,1,0,0],[0,0,0,0,1,0],[0,0,0,0,0,1]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([[1,1,1,1,0,0]])
        static_behavior_nonlinear_eq_indices=[0]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*(Ca*abs(x[0])+Cr*(x[1]**2+x[2]**2)**0.5+Cw*w[0])+x[3] if w[0]!=0 else x[3]]
        kinematic_matrix=npy.array([[1],[0],[0],[0],[0],[0]])
        static_require_kinematic=True
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
                                  static_behavior_nonlinear_eq,kinematic_matrix,
                                  static_require_kinematic,name)

    def ChangeCoefficients(self,Ca,Cr,Cw):
        self.Ca=Ca
        self.Cr=Cr
        self.Cw=Cw
        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*(Ca*abs(x[0])+Cr*(x[1]**2+x[2]**2)**0.5+Cw*w[0])+x[3] if w[0]!=0 else x[3]]
            
            

#class CylindricalLinkage(Linkage):
#    def __init__(self,part1,part2,position,euler_angles,name=''):
#        static_matrix=npy.array([[0,0,0,0],[1,0,0,0],[0,1,0,0],[0,0,0,0],[0,0,1,0],[0,0,0,1]])
#        static_linear_eq_indices=[True,True,True,True,True]
#        static_linear_eq=npy.array([[1,0,0,0,0],[0,1,0,0,0],[0,0,1,0,0],[0,0,0,0,0],[0,0,0,1,0],[0,0,0,0,1]])
#        nonlinear_static_eq=[]
#        Linkage.__init__(self,part1,part2,position,euler_angles,static_matrix,static_behavior_occurence_matrix,static_behavior_linear_eq_indices,static_behavior_linear_eq,static_behavior_nonlinear_eq,name)
#
#class PrismaticLinkage(Linkage):
#    def __init__(self,part1,part2,position,euler_angles,name=''):
#        Linkage.__init__(self,part1,part2,position,euler_angles,static_matrix,static_behavior_occurence_matrix,static_behavior_linear_eq_indices,static_behavior_linear_eq,static_behavior_nonlinear_eq,name)
#

class FrictionlessBallLinkage(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,name=''):
        static_matrix2=npy.array([[1,0,0],[0,1,0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([])
        static_behavior_nonlinear_eq_indices=[]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[]
        kinematic_matrix=npy.array([[1,0,0],[0,1,0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
        static_require_kinematic=False
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
                                  static_behavior_nonlinear_eq,kinematic_matrix,static_require_kinematic,name)
#
class FrictionlessLinearAnnularLinkage(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,name='Frictionless Linear Annular Linkage'):
        static_matrix2=npy.array([[0,0],[1,0],[0,1],[0,0],[0,0],[0,0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([])
        static_behavior_nonlinear_eq_indices=[]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[]
        kinematic_matrix=npy.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1],[0,0,0,0],[0,0,0,0]])
        static_require_kinematic=False
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,
                                  static_behavior_linear_eq,static_behavior_nonlinear_eq,
                                  kinematic_matrix,static_require_kinematic,name)



class BallLinkage(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,Ca,Cr,Cw,name='Ball Linkage'):
        self.Ca=Ca
        self.Cr=Cr
        self.Cw=Cw
        static_matrix2=npy.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1],[0,0,0,0],[0,0,0,0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([[1,1,1,1]])
        static_behavior_nonlinear_eq_indices=[0]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*(Ca*abs(x[0])+Cr*(x[1]**2+x[2]**2)**0.5+Cw*w[0])+x[3] if w[0]!=0 else x[3]]
        kinematic_matrix=npy.array([[1,0,0],[0,1,0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
        static_require_kinematic=True
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
                                  static_behavior_nonlinear_eq,kinematic_matrix,
                                  static_require_kinematic,name)

    def ChangeCoefficients(self,Ca,Cr,Cw):
        self.Ca=Ca
        self.Cr=Cr
        self.Cw=Cw
        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*(Ca*abs(x[0])+Cr*(x[1]**2+x[2]**2)**0.5+Cw*w[0])+x[3] if w[0]!=0 else x[3]]
            
    

class LinearAnnularLinkage(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,Cr,Cw,name='Linear Annular Linkage'):
        self.Cr=Cr
        self.Cw=Cw
        static_matrix2=npy.array([[0,0,0],[1,0,0],[0,1,0],[0,0,1],[0,0,0],[0,0,0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([[1,1,1]])
        static_behavior_nonlinear_eq_indices=[0]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*(Cr*(w[0]**2+x[1]**2)**0.5+Cw*w[0])+x[2] if w[0]!=0 else x[2]]
        kinematic_matrix=npy.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1],[0,0,0,0],[0,0,0,0]])
        static_require_kinematic=True
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
                                  static_behavior_nonlinear_eq,kinematic_matrix,
                                  static_require_kinematic,name)
        
        
    def ChangeCoefficients(self,Cr,Cw):
        self.Cr=Cr
        self.Cw=Cw
        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*(Cr*(w[0]**2+x[1]**2)**0.5+Cw*w[0])+x[2] if w[0]!=0 else x[2]]

#class AxialStop(HolonomicLinkage):
#    def __init__(self,part1,part2,position,euler_angles,Ca,Cw,name='Axial Stop'):
#        self.Ca=Ca
#        self.Cw=Cw
#        static_matrix2=npy.array([[1,0],[0,0],[0,0],[0,1],[0,0],[0,0]])
#        static_matrix1=-static_matrix2
#        static_behavior_occurence_matrix=npy.array([[1,1]])
#        static_behavior_nonlinear_eq_indices=[0]
#        static_behavior_linear_eq=npy.array([])
#        static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*Ca*abs(x[0])+x[1] if w[0]!=0 else x[1]]
#        kinematic_matrix=npy.array([[1,0,0,0,0],[0,1,0,0,0],[0,0,1,0,0],[0,0,0,0,0],[0,0,0,1,0],[0,0,0,0,1]])
#        static_require_kinematic=True
#        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
#                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
#                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
#                                  static_behavior_nonlinear_eq,kinematic_matrix,
#                                  static_require_kinematic,name)
#        
#        
#    def ChangeCoefficients(self,Ca,Cw):
#        self.Ca=Ca
#        self.Cw=Cw
#        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(w[0])/w[0]*Ca*abs(x[0])+x[1] if w[0]!=0 else x[1]]


class RotationalStop(HolonomicLinkage):
    def __init__(self,part1,part2,position,euler_angles,name='Rotational Stop'):
        static_matrix2=npy.array([[0],[0],[0],[1],[0],[0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([])
        static_behavior_nonlinear_eq_indices=[]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[]
        kinematic_matrix=npy.array([[0,0,0,0,0],[1,0,0,0,0],[0,1,0,0,0],[0,0,1,0,0],[0,0,0,1,0],[0,0,0,0,1]])
        static_require_kinematic=True
        HolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                  static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                  static_behavior_nonlinear_eq_indices,static_behavior_linear_eq,
                                  static_behavior_nonlinear_eq,kinematic_matrix,
                                  static_require_kinematic,name)
   
class FrictionLessGearSetLinkage(NonHolonomicLinkage):
    """
    :param alpha: pressure angle
    :param beta: helix angle
    """
    def __init__(self,part1,part2,position,euler_angles,alpha,beta,name='Gear Set Linkage'):
        self.alpha=alpha# pressure angle
        self.beta=beta# hélix angle
        
        static_matrix2=npy.array([[cos(beta)*cos(alpha),0],[sin(beta),0],[0,-1],[0,0],[0,0],[0,0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([[1,1]])
        static_behavior_nonlinear_eq_indices=[0]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[lambda x,w,v:abs(sin(beta)*cos(alpha)*abs(x[0]))+x[1]]
        directions=[npy.array([1,0,0])]
        static_require_kinematic=True
        NonHolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                     static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                     static_behavior_nonlinear_eq_indices,
                                     static_behavior_linear_eq,static_behavior_nonlinear_eq,
                                     directions,static_require_kinematic,name)
        

    def ChangeParameters(self,alpha,beta):
        self.alpha=alpha
        self.beta=beta
        self.static_matrix2=npy.array([[cos(beta)*cos(alpha),0,0],[sin(beta),0,0],[0,0,-1],[0,0,0],[0,0,0],[0,0,0]])
        self.static_matrix1=npy.array([[0,cos(beta)*cos(alpha),0],[0,sin(beta),0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(sin(beta)*cos(alpha)*abs(x[0]))+x[1]]

class GearSetLinkage(NonHolonomicLinkage):
    """
    :param alpha: pressure angle
    :param beta: helix angle
    """
    def __init__(self,part1,part2,position,euler_angles,alpha,beta,Cf,Cv,name='Gear Set Linkage'):
        self.alpha=alpha# pressure angle
        self.beta=beta# hélix angle
        self.Cf=Cf# force coefficient
        self.Cv=Cv# Speed coefficient
        
        static_matrix2=npy.array([[cos(beta)*cos(alpha),0,0],[sin(beta),0,0],[0,0,-1],[0,0,0],[0,0,0],[0,0,0]])
        static_matrix1=npy.array([[0,cos(beta)*cos(alpha),0],[0,sin(beta),0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
        static_behavior_occurence_matrix=npy.array([[1,1,1],[1,1,0]])
        static_behavior_nonlinear_eq_indices=[0,1]
        static_behavior_linear_eq=npy.array([])
#        static_behavior_nonlinear_eq=[lambda x,w,v:abs(sin(beta)*cos(alpha)*max(abs(x[0]),abs(x[1])))+x[2],
#                                      lambda x,w,v: x[1]-x[0]*(Cf*(1+sin(beta)**2*cos(alpha)**2)**0.5-1)+Cv*abs(v[0])
#                                      if v[0]*x[0]>0 else x[0]-x[1]*(Cf*(1+sin(beta)**2*cos(alpha)**2)**0.5-1)+Cv*abs(v[0])]
        static_behavior_nonlinear_eq=self.UpdateBehavior()

        directions=[npy.array([1,0,0])]
        static_require_kinematic=True
        NonHolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                     static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                     static_behavior_nonlinear_eq_indices,
                                     static_behavior_linear_eq,static_behavior_nonlinear_eq,
                                     directions,static_require_kinematic,name)


    def UpdateBehavior(self):
        f1=lambda x,w,v:abs(sin(self.beta)*cos(self.alpha)*max(abs(x[0]),abs(x[1])))+x[2]
        
        def f2(x,w,v):
            if x[0]==0.:
                if v[0]*x[1]>0:
                    r=x[1]-x[0]*(self.Cf*(1+sin(self.beta)**2*cos(self.alpha)**2)**0.5-1)+self.Cv*abs(v[0])
                else:
                    r=x[0]-x[1]*(self.Cf*(1+sin(self.beta)**2*cos(self.alpha)**2)**0.5-1)+self.Cv*abs(v[0])
            else:                
                if v[0]*x[0]<0:
                    r=x[1]-x[0]*(self.Cf*(1+sin(self.beta)**2*cos(self.alpha)**2)**0.5-1)+self.Cv*abs(v[0])
                else:
                    r=x[0]-x[1]*(self.Cf*(1+sin(self.beta)**2*cos(self.alpha)**2)**0.5-1)+self.Cv*abs(v[0])
            return r
        
        return [f1,f2]
        
        
    def ChangeCoefficients(self,Cf,Cv):
        self.Cf=Cf
        self.Cv=Cv
        
        self.static_behavior_nonlinear_eq=self.UpdateBehavior()
        
    def ChangeParameters(self,alpha,beta):
        self.alpha=alpha
        self.beta=beta
        self.static_matrix2=npy.array([[cos(beta)*cos(alpha),0,0],[sin(beta),0,0],[0,0,-1],[0,0,0],[0,0,0],[0,0,0]])
        self.static_matrix1=npy.array([[0,cos(beta)*cos(alpha),0],[0,sin(beta),0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
        
        self.static_behavior_nonlinear_eq=self.UpdateBehavior()

class FrictionlessGearSetLinkage(NonHolonomicLinkage):
    """
    :param alpha: pressure angle
    :param beta: helix angle
    """
    def __init__(self,part1,part2,position,euler_angles,alpha,beta,name='Gear Set Linkage'):
        self.alpha=alpha# pressure angle
        self.beta=beta# hélix angle
        
        static_matrix2=npy.array([[cos(beta)*cos(alpha),0],[sin(beta),0],[0,-1],[0,0],[0,0],[0,0]])
        static_matrix1=-static_matrix2
        static_behavior_occurence_matrix=npy.array([[1,1,1]])
        static_behavior_nonlinear_eq_indices=[0]
        static_behavior_linear_eq=npy.array([])
        static_behavior_nonlinear_eq=[lambda x,w,v:abs(sin(beta)*cos(alpha)*abs(x[0]))+x[1]]
        directions=[npy.array([1,0,0])]
        static_require_kinematic=True
        NonHolonomicLinkage.__init__(self,part1,part2,position,euler_angles,
                                     static_matrix1,static_matrix2,static_behavior_occurence_matrix,
                                     static_behavior_nonlinear_eq_indices,
                                     static_behavior_linear_eq,static_behavior_nonlinear_eq,
                                     directions,static_require_kinematic,name)
        
#    def ChangeCoefficients(self,Cf,Cv):
#        self.Cf=Cf
#        self.Cv=Cv
#        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(sin(self.beta)*cos(self.alpha)*max(abs(x[0]),abs(x[1])))+x[2],
#                                      lambda x,w,v: x[1]-x[0]*(Cf*(1+sin(self.beta)**2*cos(self.alpha)**2)**0.5-1)+Cv*abs(v[0])
#                                      if v[0]*x[0]>0 else x[0]-x[1]*(Cf*(1+sin(self.beta)**2*cos(self.alpha)**2)**0.5-1)+Cv*abs(v[0])]

#    def ChangeParameters(self,alpha,beta):
#        self.alpha=alpha
#        self.beta=beta
#        self.static_matrix2=npy.array([[cos(beta)*cos(alpha),0,0],[sin(beta),0,0],[0,0,-1],[0,0,0],[0,0,0],[0,0,0]])
#        self.static_matrix1=npy.array([[0,cos(beta)*cos(alpha),0],[0,sin(beta),0],[0,0,1],[0,0,0],[0,0,0],[0,0,0]])
#        self.static_behavior_nonlinear_eq=[lambda x,w,v:abs(sin(beta)*cos(alpha)*max(abs(x[0]),abs(x[1])))+x[2],
#                                      lambda x,w,v: x[1]-x[0]*(self.Cf*(1+sin(beta)**2*cos(alpha)**2)**0.5-1)+self.Cv*abs(v[0])
#                                      if v[0]*x[0]>0 else x[0]-x[1]*(self.Cf*(1+sin(beta)**2*cos(alpha)**2)**0.5-1)+self.Cv*abs(v[0])]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""


"""
import os
import webbrowser

import math
import random
import numpy as npy
import cma

from matplotlib.colors import hsv_to_rgb
import matplotlib.pyplot as plt

from matplotlib.patches import Arrow

import networkx as nx

from jinja2 import Environment, PackageLoader, select_autoescape

from dessia_common.core import DessiaObject
#from numpy import zeros
from scipy.optimize import minimize
import volmdlr as vm
from genmechanics.core import Part, Mechanism


class Parameter(DessiaObject):
    def __init__(self, lower_bound, upper_bound, periodicity=None):
        DessiaObject.__init__(self,
                              lower_bound=lower_bound,
                              upper_bound=upper_bound,
                              periodicity=periodicity)
    def random_value(self):
        return random.uniform(self.lower_bound, self.upper_bound)

    def are_values_equal(self, value1, value2, tol=1e-3):
        if self.periodicity is not None:
            value1 = value1 % self.periodicity
            value2 = value2 % self.periodicity

        return math.isclose(value1, value2, abs_tol=tol)


    def optimizer_bounds(self):
        if self.periodicity is not None:
            return (self.lower_bound-0.5*self.periodicity,
                    self.upper_bound+0.5*self.periodicity)
        else:
            return (self.lower_bound, self.upper_bound)


class Linkage(DessiaObject):
    _non_serializable_attributes = ['part1_position_function',
                                    'part2_position_function',
                                    'part1_basis_function',
                                    'part2_basis_function']
    
    
    def __init__(self,
                 part1, part1_position_function, part1_basis_function,
                 part2, part2_position_function, part2_basis_function,
                 positions_require_kinematic_parameters,
                 basis_require_kinematic_parameters,
                 kinematic_parameters,
                 name=''):
        """

        """
        DessiaObject.__init__(self,
                              part1=part1,
                              part1_position_function=part1_position_function,
                              part1_basis_function=part1_basis_function,
                              part2=part2,
                              part2_position_function=part2_position_function,
                              part2_basis_function=part2_basis_function,
                              positions_require_kinematic_parameters=positions_require_kinematic_parameters,
                              basis_require_kinematic_parameters=basis_require_kinematic_parameters,
                              kinematic_parameters=kinematic_parameters,
                              number_kinematic_parameters=len(kinematic_parameters),
                              name=name)
                
    def equivalence_hash(self):
        h = 0
        if hasattr(self, 'part1_position'):
            h += hash(self.part1_position)
            
        if hasattr(self, 'part2_position'):
            h += hash(self.part2_position)
            
        if hasattr(self, 'part1_basis'):
            h += hash(self.part1_basis)
            
        if hasattr(self, 'part2_basis'):
            h += hash(self.part2_basis)
            
        return h
        
    def is_equivalent(self, other_linkage):
        if self.__class__ != other_linkage.__class__:
            return False

        if hasattr(self, 'part1_position'):
            if self.part1_position != other_linkage.part1_position:
                return False
            
        if hasattr(self, 'part2_position'):
            if self.part2_position != other_linkage.part2_position:
                return False
            
        if hasattr(self, 'part1_basis'):
            if self.part1_basis != other_linkage.part1_basis:
                return False
            
        if hasattr(self, 'part2_basis'):
            if self.part2_basis != other_linkage.part2_basis:
                return False
            
        return True
        


    def frame(self, linkage_parameters_values, side):
        if side:

            part1_frame = self.part1_basis_function(linkage_parameters_values)\
                .to_frame(self.part1_position_function(linkage_parameters_values))

            part2_frame = -self.part2_basis_function(linkage_parameters_values)\
                .to_frame(-self.part2_position_function(linkage_parameters_values))

            return part1_frame + part2_frame

        else:
            part1_frame = -self.part1_basis_function(linkage_parameters_values)\
                .to_frame(-self.part1_position_function(linkage_parameters_values))

            part2_frame = self.part2_basis_function(linkage_parameters_values)\
                .to_frame(self.part2_position_function(linkage_parameters_values))

            return part2_frame + part1_frame

    def babylonjs(self, initial_linkage_parameters,
                  part1_parent=None, part2_parent=None):
        part1_position = self.part1_position_function(initial_linkage_parameters)
        part2_position = self.part2_position_function(initial_linkage_parameters)

        s = ''
        if part1_parent is not None:
            s += 'var linkage_part1_mesh = BABYLON.MeshBuilder.CreateSphere("default_linkage part1", {diameter: 0.02}, scene);\n'
            s += 'linkage_part1_mesh.position = new BABYLON.Vector3({}, {}, {});\n'.format(*part1_position.vector)
            s += 'linkage_part1_mesh.parent = {};\n'.format(part1_parent)

        if part2_parent:
            s += 'var linkage_part2_mesh = BABYLON.MeshBuilder.CreateSphere("default_linkage part2", {diameter: 0.02}, scene);\n'
            s += 'linkage_part2_mesh.position = new BABYLON.Vector3({}, {}, {});\n'.format(*part2_position.vector)
            s += 'linkage_part2_mesh.parent = {};\n'.format(part2_parent)

        return s


class RevoluteLinkage(Linkage):
    holonomic = True

    def __init__(self,
                 part1, part1_position, part1_basis,
                 part2, part2_position, part2_basis, name='RevoluteLinkage'):
        """
        :param part2_basis: a basis defining orientation of linkage on part2

        """

        def part1_basis_f(q):
            return part1_basis.rotation(part1_basis.u, q[0], copy=True)

        def part2_basis_f(q):
            return part2_basis

        DessiaObject.__init__(self,
                              part1_position=part1_position,
                              part2_position=part2_position,
                              part1_basis=part1_basis,
                              part2_basis=part2_basis)

        Linkage.__init__(self,
                         part1, lambda q: part1_position, part1_basis_f,
                         part2, lambda q: part2_position, part2_basis_f,
                         False, True,
                         [Parameter(0., 2*math.pi, 2*math.pi)], name=name)



    def babylonjs(self, initial_linkage_parameters,
                  part1_parent=None, part2_parent=None):
        s = ''
        if part1_parent is not None:

            s += 'var path1 = [new BABYLON.Vector3({}, {}, {}), new BABYLON.Vector3({}, {}, {})];\n'.format(*(self.part1_position-0.03*self.part1_basis.u),
                                                                                                           *(self.part1_position+0.03*self.part1_basis.u))
            s += 'var linkage_part1_mesh = BABYLON.MeshBuilder.CreateTube("revolute part1", {path: path1, radius: 0.01, sideOrientation:BABYLON.Mesh.DOUBLESIDE}, scene);\n'
            s += 'linkage_part1_mesh.enableEdgesRendering();\n'
            s += 'linkage_part1_mesh.edgesWidth = 0.4;\n'
            s += 'linkage_part1_mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 1);\n'
            s += 'linkage_part1_mesh.parent = {};\n'.format(part1_parent)


        if part2_parent is not None:
            s += 'var path2 = [new BABYLON.Vector3({}, {}, {}), new BABYLON.Vector3({}, {}, {})];\n'.format(*(self.part2_position-0.03*self.part2_basis.u),
                                                                                                           *(self.part2_position+0.03*self.part2_basis.u))
            s += 'var linkage_part2_mesh = BABYLON.MeshBuilder.CreateTube("revolute part2", {path: path2, radius: 0.015, sideOrientation:BABYLON.Mesh.DOUBLESIDE}, scene);\n'
            s += 'linkage_part2_mesh.enableEdgesRendering();\n'
            s += 'linkage_part2_mesh.edgesWidth = 0.4;\n'
            s += 'linkage_part2_mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 1);\n'
            s += 'linkage_part2_mesh.parent = {};\n'.format(part2_parent)

        return s


class SlidingRevoluteLinkage(Linkage):
    holonomic = True

    def __init__(self,
                 part1, part1_position, part1_basis,
                 part2, part2_position, part2_basis, name='SlidingRevoluteLinkage'):
        """
        :param part2_basis: a basis defining orientation of linkage on part2
        The first kineamtic parameter is the translation, the second the rotation
        """

        def part1_position_f(q):
            return part1_position + q[0]*part1_basis.u

        def part2_position_f(q):
            return part2_position

        def part1_basis_f(q):
            return part1_basis.Rotation(part1_basis.u, q[1], copy=True)

        DessiaObject.__init__(self,
                              part1_position=part1_position,
                              part2_position=part2_position,
                              part1_basis=part1_basis,
                              part2_basis=part2_basis)

        Linkage.__init__(self,
                         part1, part1_position_f, part1_basis_f,
                         part2, part2_position_f, lambda q: part2_basis,
                         True, True,
                         [Parameter(0., 2*math.pi, 2*math.pi),
                          Parameter(-1., 1., None)], name)



    def babylonjs(self, initial_linkage_parameters,
                  part1_parent=None, part2_parent=None):
#        part1_position = self.part1_position_function(initial_linkage_parameters)
#        part1_basis = self.part1_position(initial_linkage_parameters)

#        part2_position = self.part2_position_function(initial_linkage_parameters)
#        part2_basis = self.part2_position(initial_linkage_parameters)
        s = ''
        if part1_parent is not None:

            s += 'var path1 = [new BABYLON.Vector3({}, {}, {}), new BABYLON.Vector3({}, {}, {})];\n'.format(*(self.part1_position-0.1*self.part1_basis.u),
                                                                                                            *(self.part1_position+0.1*self.part1_basis.u))
            s += 'var linkage_part1_mesh = BABYLON.MeshBuilder.CreateTube("revolute part1", {path: path1, radius: 0.01, sideOrientation:BABYLON.Mesh.DOUBLESIDE}, scene);\n'
            s += 'linkage_part1_mesh.enableEdgesRendering();\n'
            s += 'linkage_part1_mesh.edgesWidth = 0.4;\n'
            s += 'linkage_part1_mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 1);\n'
            s += 'linkage_part1_mesh.parent = {};\n'.format(part1_parent)


        if part2_parent is not None:
            s += 'var path2 = [new BABYLON.Vector3({}, {}, {}), new BABYLON.Vector3({}, {}, {})];\n'.format(*(self.part2_position-0.03*self.part2_basis.u),
                                                                                                           *(self.part2_position+0.03*self.part2_basis.u))
            s += 'var linkage_part2_mesh = BABYLON.MeshBuilder.CreateTube("revolute part2", {path: path2, radius: 0.015, sideOrientation:BABYLON.Mesh.DOUBLESIDE}, scene);\n'
            s += 'linkage_part2_mesh.enableEdgesRendering();\n'
            s += 'linkage_part2_mesh.edgesWidth = 0.4;\n'
            s += 'linkage_part2_mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 1);\n'
            s += 'linkage_part2_mesh.parent = {};\n'.format(part2_parent)

        return s

class PrismaticLinkage(Linkage):
    holonomic = True


    def __init__(self,
                 part1, part1_position, part1_basis,
                 part2, part2_position, part2_basis, name='PrismaticLinkage'):
        """
        :param part2_basis: a basis defining orientation of linkage on part2

        """

        def part1_position_f(q):
            return part1_position + q[0]*part1_basis.u

        def part2_position_f(q):
            return part2_position

        DessiaObject.__init__(self,
                              part1_position=part1_position,
                              part2_position=part2_position,
                              part1_basis=part1_basis,
                              part2_basis=part2_basis)

        Linkage.__init__(self,
                         part1, part1_position_f, lambda q: part1_basis,
                         part2, part2_position_f, lambda q: part2_basis,
                         True, False,
                         [Parameter(-1, 1, None)], name)




#    def __init__(self,
#                 part1, part1_initial_position, part1_axis,
#                 part2, part2_position, part2_basis=vm.xyz.copy(),
#                 name=''):
#        """
#        sliding on part1, fixed on part2
#        """
#
#        def basis(q):
#            return part2_basis
#
#        Linkage.__init__(self,
#                         part1, lambda q: part1_initial_position+q[0]*part1_axis,
#                         part2, lambda q: part2_position,
#                         True, False,
#                         basis, 1, name)
#
#        DessiaObject.__init__(self,
#                              part1_axis=part1_axis,
#                              part2_basis=part2_basis)

    def babylonjs(self, initial_linkage_parameters, part1_parent=None, part2_parent=None):

        bp1 = self.part1_basis_function(initial_linkage_parameters)
        bp2 = self.part2_basis_function(initial_linkage_parameters)
        s = ''
        if part1_parent is not None:
            s += 'var linkage_part1_mesh = BABYLON.MeshBuilder.CreateBox("prismatic part1", {depth:0.015, height:0.015, width:0.25}, scene);\n'
            s += 'linkage_part1_mesh.parent = {};\n'.format(part1_parent)
            s += 'linkage_part1_mesh.position = new BABYLON.Vector3({}, {}, {});\n'.format(*self.part1_position_function(initial_linkage_parameters))
            s += 'linkage_part1_mesh.rotation = BABYLON.Vector3.RotationFromAxis(new BABYLON.Vector3({}, {}, {}),new BABYLON.Vector3({}, {}, {}), new BABYLON.Vector3({}, {}, {}));\n'.format(*bp1.u, *bp1.v, *bp1.w)
            s += 'linkage_part1_mesh.enableEdgesRendering();\n'
            s += 'linkage_part1_mesh.edgesWidth = 0.3;\n'
            s += 'linkage_part1_mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 1);\n'

        if part2_parent is not None:
            s += 'var linkage_part2_mesh = BABYLON.MeshBuilder.CreateBox("prismatic part2", {depth:0.03, height:0.03, width:0.06}, scene);\n'
            s += 'linkage_part2_mesh.parent = {};\n'.format(part2_parent)
            s += 'linkage_part2_mesh.position = new BABYLON.Vector3({}, {}, {});\n'.format(*self.part2_position_function(initial_linkage_parameters))
            s += 'linkage_part2_mesh.rotation = BABYLON.Vector3.RotationFromAxis(new BABYLON.Vector3({}, {}, {}),new BABYLON.Vector3({}, {}, {}), new BABYLON.Vector3({}, {}, {}));\n'.format(*bp2.u, *bp2.v, *bp2.w)
            s += 'linkage_part2_mesh.enableEdgesRendering();\n'
            s += 'linkage_part2_mesh.edgesWidth = 0.3;\n'
            s += 'linkage_part2_mesh.edgesColor = new BABYLON.Color4(0, 0, 0, 1);\n'

        return s

class LimitedBallLinkage(Linkage):
    holonomic = True

    def __init__(self,
                 part1, part1_position, part1_basis,
                 part2, part2_position, part2_basis,
                 name='LimitedBallLinkage'):
        """
        Allowed movements are:
            - a rotation around part1 basis u
            - a rotation around part1 basis v
            
        """

        def part1_basis_f(q):
            return part1_basis.Rotation(part1_basis.u, q[0], copy=True)\
                              .Rotation(part1_basis.v, q[1], copy=True)

        def part2_basis_f(q):
            return part2_basis


        DessiaObject.__init__(self,
                              part1_position=part1_position,
                              part2_position=part2_position,
                              part1_basis=part1_basis,
                              part2_basis=part2_basis)

        Linkage.__init__(self,
                         part1, lambda q: part1_position, part1_basis_f,
                         part2, lambda q: part2_position, part2_basis_f,
                         False, True,
                         [Parameter(0., 2*math.pi, 2*math.pi),
                          Parameter(0., 2*math.pi, 2*math.pi)],
                         name)

class BallLinkage(Linkage):
    holonomic = True

    def __init__(self,
                 part1, part1_position, part1_basis,
                 part2, part2_position, part2_basis,
                 name='BallLinkage'):
        """

        """

        DessiaObject.__init__(self,
                              part1_position=part1_position,
                              part2_position=part2_position,
                              part1_basis=part1_basis,
                              part2_basis=part2_basis)

        part1_basis_f, part2_basis_f = self.basis_functions()

        Linkage.__init__(self,
                         part1, lambda q: part1_position, part1_basis_f,
                         part2, lambda q: part2_position, part2_basis_f,
                         False, True,
                         [Parameter(0., 2*math.pi, 2*math.pi),
                          Parameter(0., 2*math.pi, 2*math.pi),
                          Parameter(0., 2*math.pi, 2*math.pi)], name)


    def update_part1_point(self, new_position):
        self.part1_position = new_position
        self.part1_position_function = lambda q: new_position


    def update_part2_point(self, new_position):
        self.part2_position = new_position
        self.part2_position_function = lambda q: new_position

    def basis_functions(self):
        
        def part1_basis_f(q):
            return self.part1_basis.Rotation(self.part1_basis.u, q[0], copy=True)\
                              .Rotation(self.part1_basis.v, q[1], copy=True)\
                              .Rotation(self.part1_basis.w, q[2], copy=True)

        def part2_basis_f(q):
            return self.part2_basis
        
        return part1_basis_f, part2_basis_f

class NoConfigurationFoundError(Exception):
    pass

class MovingMechanism(Mechanism):
    def __init__(self, linkages, ground, name=''):

        Mechanism.__init__(self,
                           linkages,
                           ground,
                           {},
                           None,
                           None,
                           name=name)

        # self.parts_setting_path = {}
        self._settings_path = {}

        # Settings
        self.settings_graph()
        n_kp = 0
        self.kinematic_parameters_mapping = {}
        for linkage in self.linkages_kinematic_setting:
            for i in range(linkage.number_kinematic_parameters):
                self.kinematic_parameters_mapping[linkage, i] = n_kp + i
            n_kp += linkage.number_kinematic_parameters

    def settings_graph(self):
        graph = self.holonomic_graph.copy()
        self.opened_linkages = []
        # print('cycles deleting')
        # graph_has_cycles = True
        graph_cycles = nx.cycle_basis(graph)
        
        while len(graph_cycles) != 0:
            # Deleting first cycle of graph
            
            ground_distance = [(l, len(nx.shortest_path(graph, l, self.ground)))\
                               for l in graph_cycles[0]\
                               if l in self.linkages\
                                   and not l in self.opened_linkages\
                                   and not l.positions_require_kinematic_parameters
                               ]

            linkage_to_delete = max(ground_distance, key=lambda x:x[1])[0]
            # print(linkage_to_delete.name)
            self.opened_linkages.append(linkage_to_delete)
            graph.remove_node(linkage_to_delete)
            graph_cycles = nx.cycle_basis(graph)

        self.linkages_kinematic_setting = [l for l in self.linkages if l not in self.opened_linkages]
        self.settings_graph = graph


    def plot_settings_graph(self):
        s="""<html>
        <head>
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.20.0/vis.min.js"></script>
        <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.20.0/vis.min.css" rel="stylesheet" type="text/css" />

        <style type="text/css">
            #mynetwork {
                border: 1px solid lightgray;
            }
        </style>
    </head>
    <body>
    <div id="mynetwork"></div>

    <script type="text/javascript">
    var nodes = new vis.DataSet([\n"""
        index={}
        for ipart,part in enumerate(self.parts+[self.ground]):
            index[part]=ipart
            s+="{{id: {}, label: '{}'}},\n".format(ipart,part.name)
#        s+=']);\n'
        n=len(self.parts)+1
#        index[self.ground]=n
#        n+=1
        for il,linkage in enumerate(self.linkages_kinematic_setting):
            index[linkage] = n+il
            s+="{{id: {}, label: '{}'}},\n".format(n+il,linkage.name)
        s+=']);\n'

        s+="var edges = new vis.DataSet(["
        for linkage in self.linkages_kinematic_setting:
            s+='{{from: {}, to: {}}},\n'.format(index[linkage],index[linkage.part1])
            s+='{{from: {}, to: {}}},\n'.format(index[linkage],index[linkage.part2])
        s+=']);'

        s+="""
    // create a network
    var container = document.getElementById('mynetwork');

    // provide the data in the vis format
    var data = {
        nodes: nodes,
        edges: edges
    };
    var options = {};

    // initialize your network!
    var network = new vis.Network(container, data, options);
</script>
</body>
</html>"""

        with open('gm_graph_viz.html','w') as file:
            file.write(s)

        webbrowser.open('file://' + os.path.realpath('gm_graph_viz.html'))


    def settings_path(self, part1, part2):
        if (part1, part2) in self._settings_path:
            return self._settings_path[part1, part2]
        elif (part2, part1) in self._settings_path:
            path = [(p2, linkage, not linkage_side, p1) for (p1, linkage, linkage_side, p2) in self._settings_path[part2, part1][::-1]]
            self._settings_path[part1, part2] = path
            return self._settings_path[part1, part2]
        else:
            path = []
            try:
                raw_path = list(nx.shortest_path(self.settings_graph, part1, part2))
            except nx.NetworkXNoPath:
                self.plot_settings_graph()
                print(part1.name, part2.name)
                raise nx.NetworkXError
            for path_part1, linkage, path_part2 in zip(raw_path[:-2:2], raw_path[1::2], raw_path[2::2]+[part2]):
                path.append((path_part1, linkage, linkage.part1==path_part1, path_part2))

            self._settings_path[part1, part2] = path
            return path

    def part_global_frame(self, part, kinematic_parameters_values):
        frame = vm.OXYZ
        for part1, linkage, linkage_side, part2 in self.settings_path(self.ground, part):
            
            linkage_parameters_values = self.extract_linkage_parameters_values(linkage, kinematic_parameters_values)
            linkage_frame = linkage.frame(linkage_parameters_values, side=linkage_side)
            frame = frame + linkage_frame

        return frame


    def part_relative_frame(self, part, reference_part, kinematic_parameters_values):
        frame = vm.OXYZ
        for part1, linkage, linkage_side, part2 in self.settings_path(reference_part, part):
            linkage_parameters_values = self.extract_linkage_parameters_values(linkage, kinematic_parameters_values)
            linkage_frame = linkage.frame(linkage_parameters_values, side=linkage_side)
            frame = frame + linkage_frame

        return frame    

    def linkage_global_position(self, linkage, global_parameter_values):
        if linkage.positions_require_kinematic_parameters:
            ql = self.extract_linkage_parameters_values(linkage,
                                                        global_parameter_values)
        else:
            ql = []


        part1_frame = self.part_global_frame(linkage.part1, global_parameter_values)
        return part1_frame.OldCoordinates(linkage.part1_position_function(ql))


    def extract_linkage_parameters_values(self, linkage, global_parameter_values):
      
        linkage_parameters = [global_parameter_values[self.kinematic_parameters_mapping[linkage, i]]\
               for i in range(linkage.number_kinematic_parameters)]
        return linkage_parameters

    def global_to_linkages_parameter_values(self, global_parameter_values):
        linkages_parameter_values = {}
        for linkage in self.linkages_kinematic_setting:
            linkages_parameter_values[linkage] = self.extract_linkage_parameters_values(linkage, global_parameter_values)
        return linkages_parameter_values
            

    def opened_linkage_gap(self, linkage, global_parameter_values):
        if linkage.positions_require_kinematic_parameters:
            ql = self.extract_linkage_parameters_values(linkage, global_parameter_values)
        else:
            ql = []
        position1 = self.part_global_frame(linkage.part1, global_parameter_values).old_coordinates(linkage.part1_position_function(ql))
        position2 = self.part_global_frame(linkage.part2, global_parameter_values).old_coordinates(linkage.part2_position_function(ql))
        return position2 - position1

    def opened_linkage_misalignment(self, linkage, global_parameter_values):
        ql = self.extract_linkage_parameters_values(linkage, global_parameter_values)

        basis1 = self.part_global_frame(linkage.part1, global_parameter_values).Basis()
        basis2 = self.part_global_frame(linkage.part2, global_parameter_values).Basis()
        basis = basis2 - basis1 - linkage.basis(ql)
        return basis

    def opened_linkages_residue(self, q):
        residue = 0.
        for linkage in self.opened_linkages:
            residue += self.opened_linkage_gap(linkage, q).norm()
        return residue

    def reduced_x_to_full_x(self, xr, basis_vector, free_parameters_dofs):
        x = basis_vector[:]

        for qrv, i in zip(xr, free_parameters_dofs):
            x[i] = qrv
        return x

    def full_x_to_reduced_x(self, x, free_parameters_dofs):
        return [x[i] for i in free_parameters_dofs]

    def geometric_closing_residue_function(self, basis_vector,
                                           free_parameters_dofs):


        def residue_function(xr):
            x = self.reduced_x_to_full_x(xr, basis_vector, free_parameters_dofs)


            return self.opened_linkages_residue(x)

        return residue_function


    def _optimization_settings(self, imposed_parameters):
        # Free parameter identification
        free_parameters_dofs = []
        free_parameters = []
        n_free_parameters = 0
        n_parameters = len(self.kinematic_parameters_mapping.items())
        basis_vector = [0] * n_parameters
        for i in range(n_parameters):
            if i in imposed_parameters:
                basis_vector[i] = imposed_parameters[i]
            else:
                free_parameters_dofs.append(i)
                n_free_parameters += 1

        bounds = []
        for (linkage, iparameter), idof in self.kinematic_parameters_mapping.items():
            if idof in free_parameters_dofs:
                parameter = linkage.kinematic_parameters[iparameter]
                free_parameters.append(parameter)
                bounds.append(parameter.optimizer_bounds())

            bounds_cma = [[], []]
            for bmin, bmax in bounds:
               bounds_cma[0].append(bmin)
               bounds_cma[1].append(bmax)

        return basis_vector, free_parameters_dofs, free_parameters, n_free_parameters, bounds, bounds_cma


    def find_configurations(self, imposed_parameters,
                           number_max_configurations,
                           number_starts=10, tol=1e-5,
                           starting_point=None):

        # initial_imposed_parameters = {k: v[0] for k,v in steps_imposed_parameters.items()}


        basis_vector, free_parameters_dofs, free_parameters, n_free_parameters, bounds, bounds_cma\
            = self._optimization_settings(imposed_parameters)

        geometric_closing_residue = self.geometric_closing_residue_function(basis_vector,
                                                                            free_parameters_dofs)


        # Starting n times
        starting_points = []

        for istart in range(number_starts):
            if starting_point is None:
                xr0 = [0]*n_free_parameters
                for i, parameter in enumerate(free_parameters):
                    xr0[i] = parameter.random_value()
            else:
                xr0 = [starting_point[i] for i in free_parameters_dofs]

            # result = minimize(geometric_closing_residue, xr0, bounds=bounds,
            #                   tol=0.1*tol)
            # fopt = result.fun
            # if fopt < tol:
            #     xr_opt = result.x
            # else:

            xr_opt, fopt = cma.fmin(geometric_closing_residue, xr0, 0.2,
                                  options={'bounds':bounds_cma,
                                           'ftarget': tol,
                                           'verbose': -9,
                                           'maxiter': 2000})[0:2]


            if fopt <= tol:
#                print('converged')
                found_x = False
                for x in starting_points:
                    equal = True
                    for parameter, xi1, xi2 in zip(free_parameters, x, xr_opt):
                        if not parameter.are_values_equal(xi1, xi2):
                            equal = False
                            break
                    if equal:
#                        print(x, result.x, 'are equal')
                        found_x = True

                if not found_x:
                    xopt = self.reduced_x_to_full_x(xr_opt, basis_vector, free_parameters_dofs)
                    starting_points.append(xopt[:])
                    yield xopt
                    if len(starting_points) >= number_max_configurations:
                        break

        print('Found {} configurations'.format(len(starting_points)))
        raise NoConfigurationFoundError


    def solve_from_initial_configuration(self, initial_parameter_values,
                                         steps_imposed_parameters,
                                         number_step_retries=5,
                                         max_failed_steps=3,
                                         tol=1e-4):
        """
        returns a MechanismConfigurations object. The initial point deduced from initial_parameter_values
        is the first step of the MechanismConfigurations object.
        """


        x0 = initial_parameter_values
        step_imposed_parameters = {k: v[0] for k, v in steps_imposed_parameters.items()}
        basis_vector, free_parameters_dofs, free_parameters, n_free_parameters, bounds, bounds_cma\
                = self._optimization_settings(step_imposed_parameters)
        xr0 = self.full_x_to_reduced_x(x0, free_parameters_dofs)

        n_steps = len(list(steps_imposed_parameters.values())[0])

        linkage_steps_parameters = [self.global_to_linkages_parameter_values(x0)]

        number_failed_steps = 0
        failed_step = False
        for istep in range(n_steps):
            step_imposed_parameters = {k: v[istep] for k, v in steps_imposed_parameters.items()}

            # basis vector needs update at each time step!
            basis_vector, free_parameters_dofs, free_parameters, n_free_parameters, bounds, bounds_cma\
                = self._optimization_settings(step_imposed_parameters)

            geometric_closing_residue = self.geometric_closing_residue_function(basis_vector,
                                                                                free_parameters_dofs)

            if n_free_parameters > 0:
                step_converged = False
                n_tries_step = 1
                while (not step_converged) and (n_tries_step<= number_step_retries):
                    result = minimize(geometric_closing_residue,
                                      npy.array(xr0)+0.01*(npy.random.random(n_free_parameters)-0.5),
                                      tol=0.1*tol, bounds=bounds)
                    xr_opt = result.x
                    fopt = result.fun
                    if fopt > tol:
                        xr_opt, fopt = cma.fmin(geometric_closing_residue, xr0, 0.1,
                                              options={'bounds':bounds_cma,
    #                                              'tolfun':0.5*tol,
                                                       'verbose':-9,
                                                       'ftarget': tol,
                                                       'maxiter': 500})[0:2]
                    n_tries_step += 1
                    step_converged = fopt < tol
                if step_converged:
                    xr0 = xr_opt[:]
                    x = self.reduced_x_to_full_x(xr_opt, basis_vector, free_parameters_dofs)


                    # qs.append(x[:])
                    linkage_steps_parameters.append(self.global_to_linkages_parameter_values(x))
                else:
                    print('@istep {}: residue: {}'.format(istep, fopt))
                    number_failed_steps += 1
                    if number_failed_steps >= max_failed_steps:
                        print('Failed {} steps, stopping configuration computation'.format(max_failed_steps))
                        failed_step = True
                        break


            else:
                
                linkage_steps_parameters.append(self.global_to_linkages_parameter_values(basis_vector))
                # qs.append(basis_vector)
        if not failed_step:
            return MechanismConfigurations(self,steps_imposed_parameters, linkage_steps_parameters)

#    def solve_configurations(steps_imposed_parameters,
#                             number_max_configurations)
#
#        for configuration in self.find_initial_configurations(steps_imposed_parameters,
#                                    number_max_configurations,
#                                    number_starts=10, tol=1e-5):
#
#            yield self.solve_from_initial_configuration(self, initial_parameter_values,
#                                         steps_imposed_parameters,
#                                         number_step_retries=5,
#                                         max_failed_steps=3,
#                                         tol=1e-4)

def istep_from_value_on_list(list_, value):
    for ipoint, (point1, point2) in enumerate(zip(list_[:-1],
                                                  list_[1:])):
        interval = sorted((point1, point2))
        if (interval[0] <= value) and (value <= interval[1]):
            alpha = (value-point1)/(point2-point1)
            if alpha < 0 or alpha > 1:
                raise ValueError
            return ipoint + alpha
    values = [p for p in list_]
    min_values = min(values)
    max_values = max(values)
    raise ValueError('Specified value not found in list_: {} not in [{}, {}]'.format(value, min_values, max_values))


def istep_from_value_on_trajectory(trajectory, value, axis):
    for ipoint, (point1, point2) in enumerate(zip(trajectory[:-1],
                                                  trajectory[1:])):
        interval = sorted((point1[axis], point2[axis]))
        if (interval[0] <= value) and (value <= interval[1]):
            alpha = (value-point1[axis])/(point2[axis]-point1[axis])
            if alpha < 0 or alpha > 1:
                raise ValueError
            return ipoint + alpha
    values = [p[axis] for p in trajectory]
    min_values = min(values)
    max_values = max(values)
    raise ValueError('Specified value not found in trajectory: {} not in [{}, {}]'.format(value, min_values, max_values))

def point_from_istep_on_trajectory(trajectory, istep):
    istep1 = int(istep)
    if istep1 == istep:
        # No interpolation needed
        return trajectory[int(istep)]
    else:
        alpha = istep - istep1
        point1 = trajectory[istep1]
        point2 = trajectory[istep1+1]
        return (1-alpha)*point1+(alpha)*point2

def trajectory_point_from_value(trajectory, value, axis):
    for ipoint, (point1, point2) in enumerate(zip(trajectory[:-1],
                                                  trajectory[1:])):
        interval = sorted((point1[axis], point2[axis]))
        if (interval[0] <= value) and (value < interval[1]):
            alpha = (value - point1[axis])/(point2[axis] - point1[axis])
            return (1-alpha)*point1 + alpha*point2
    return None

def trajectory_derivative(trajectory, istep, delta_istep):
    istep1 = istep-0.5*delta_istep
    istep2 = istep+0.5*delta_istep
    if istep1 < 0:
        istep1 = 0
        istep2 = delta_istep
    if istep2 > len(trajectory)-1:
        istep2 = len(trajectory)-1
        istep1 = istep2 - delta_istep
        if istep1 < 0:
            raise ValueError('Delta istep is too large!')
    point1 = point_from_istep_on_trajectory(trajectory, istep1)
    point2 = point_from_istep_on_trajectory(trajectory, istep2)
    return (point2-point1)

class MechanismConfigurations(DessiaObject):

    def __init__(self,
                 mechanism,
                 steps_imposed_parameters,
                 linkage_steps_parameters):

        number_steps = len(linkage_steps_parameters)


        DessiaObject.__init__(self,
                              mechanism=mechanism,
                              steps_imposed_parameters=steps_imposed_parameters,
                              linkage_steps_parameters=linkage_steps_parameters,
                              number_steps=number_steps)

        if not self.is_valid():
            raise ValueError

        self.trajectories = {}

    def is_valid(self):
        # nparam_mechanism = len(self.mechanism.kinematic_parameters_mapping)
        # for istep, step in enumerate(self.steps):
        #     if len(step) != nparam_mechanism:
        #         print('Step n°{} has incorrect length'.format(istep))
        #         return False
        return True

    def opened_linkages_residue(self):
        residues = []
        for step in self.steps:
            residues.append(self.mechanism.opened_linkages_residue(step))
        return residues

    def interpolate_step(self, istep):
        """
        :param istep: can be a float
        """

        istep1 = int(istep)
        
        alpha = istep - istep1
        if alpha == 0.:
            return self.steps[istep1]

        return [(1-alpha)*s1+alpha*s2 for s1, s2 in zip(self.steps[istep1],
                                                        self.steps[istep1+1])]
        

    def plot_kinematic_parameters(self,
                                  linkage1, kinematic_parameter1,
                                  linkage2, kinematic_parameter2
                                  ):
        x = []
        y = []
        # dof1 = self.mechanism.kinematic_parameters_mapping[linkage1, kinematic_parameter1]
        # dof2 = self.mechanism.kinematic_parameters_mapping[linkage2, kinematic_parameter2]
        for step in self.linkage_steps_parameters:
            x.append(step[linkage1][kinematic_parameter1])
            y.append(step[linkage2][kinematic_parameter2])
        fig, ax = plt.subplots()
        ax.plot(x, y, marker='o')
        ax.set_xlabel('Parameter {} of linkage {}'.format(kinematic_parameter1+1, linkage1.name))
        ax.set_ylabel('Parameter {} of linkage {}'.format(kinematic_parameter2+1, linkage2.name))
        ax.grid()
        return fig, ax

    def trajectory(self, point, part, reference_part):
        if (point, part, reference_part) in self.trajectories:
            return self.trajectories[point, part, reference_part]

        trajectory = []
        for step in self.steps:
            frame1 = self.mechanism.part_global_frame(part, step)
            frame2 = self.mechanism.part_global_frame(reference_part, step)
            frame = frame1 - frame2
            trajectory.append(frame.OldCoordinates(point))

        self.trajectories[point, part, reference_part] = trajectory

        return trajectory

    def plot2D_trajectory(self, point, part, reference_part, x=vm.X3D, y=vm.Y3D, equal_aspect=True):
        xt = []
        yt = []
        for traj_point in self.trajectory(point, part, reference_part):
            xp, yp = traj_point.PlaneProjection2D(x, y)
            xt.append(xp)
            yt.append(yp)

        fig, ax = plt.subplots()
        ax.plot(xt, yt, marker='o')
        ax.grid()
        ax.set_xlabel(str(x))
        ax.set_ylabel(str(y))
        ax.set_title('Trajectory of point {} on part {} relatively to part {}'.format(str(point), part.name, reference_part.name))

        if equal_aspect:
            ax.set_aspect('equal')

        return fig, ax

    def plot_trajectory(self, point, part, reference_part, equal_aspect=True):
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        xt = []
        yt = []
        zt = []
        for point in self.trajectory(point, part, reference_part):
            xp, yp, zp = point
            xt.append(xp)
            yt.append(yp)
            zt.append(zp)

#        fig, ax = plt.subplots()
        ax.plot(xt, yt, zt, marker='o')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        ax.set_title('Trajectory of point {} on part {} relatively to part {}'.format(str(point), part.name, reference_part.name))


        if equal_aspect:
            ax.set_aspect('equal')
#        fig.canvas.set_window_title('Trajectory')
        return fig, ax



    def part_local_point_global_speed(self, part, point, istep):
        """

        """
        if (istep < 0) or (istep > self.number_steps-1):
            raise ValueError('istep outside of bounds: {}'.format(istep))
        elif istep < 0.5:
            # Backward extrapolation from speeds 1 and 2
            frame1 = self.mechanism.part_global_frame(part, self.steps[0])
            frame2 = self.mechanism.part_global_frame(part, self.steps[1])
            frame3 = self.mechanism.part_global_frame(part, self.steps[2])
            p1 = frame1.OldCoordinates(point)
            p2 = frame2.OldCoordinates(point)
            p3 = frame3.OldCoordinates(point)
            v1 = p2 - p1
            v2 = p3 - p2
            alpha = istep - 0.5
            return (1-alpha)*v1 + alpha*v2

        elif istep > self.number_steps-1.5:
            # forward extrapolation from speeds n-1 and n
            i1 = int(istep-0.5)
            frame1 = self.mechanism.part_global_frame(part, self.steps[-3])
            frame2 = self.mechanism.part_global_frame(part, self.steps[-2])
            frame3 = self.mechanism.part_global_frame(part, self.steps[-1])
            p1 = frame1.OldCoordinates(point)
            p2 = frame2.OldCoordinates(point)
            p3 = frame3.OldCoordinates(point)
            v1 = p2 - p1
            v2 = p3 - p2
            alpha = istep - (self.number_steps - 2.5)
            return (1-alpha)*v1 + alpha*v2
        else:

            int_istep = int(istep)
            if int_istep+0.5 == istep:
                # Using exact derivative
                frame1 = self.mechanism.part_global_frame(part, self.steps[int_istep])
                frame2 = self.mechanism.part_global_frame(part, self.steps[int_istep+1])
                p1 = frame1.OldCoordinates(point)
                p2 = frame2.OldCoordinates(point)
                return p2 - p1
            else:
                # interpolation in between
                i1 = int(istep-0.5)
                frame1 = self.mechanism.part_global_frame(part, self.steps[i1])
                frame2 = self.mechanism.part_global_frame(part, self.steps[i1+1])
                frame3 = self.mechanism.part_global_frame(part, self.steps[i1+2])
                p1 = frame1.OldCoordinates(point)
                p2 = frame2.OldCoordinates(point)
                p3 = frame3.OldCoordinates(point)
                v1 = p2 - p1
                v2 = p3 - p2
                alpha = istep - i1 - 0.5
                return (1-alpha)*v1 + alpha*v2



    def part_global_rotation_vector(self, part, istep):

        step = self.interpolate_step(istep)
        frame = self.mechanism.part_global_frame(part, step)

        point1 = vm.O3D
        point1_speed = self.part_local_point_global_speed(part, point1, istep)
        for point2 in [vm.X3D, vm.Y3D, vm.Z3D]:

            point2_speed = self.part_local_point_global_speed(part, point2, istep)
            delta_speeds = point2_speed - point1_speed
            if not math.isclose(delta_speeds.Norm(), 0, abs_tol=1e-8):
                break

        p21 = frame.OldCoordinates(point2) - frame.OldCoordinates(point1)
        R = delta_speeds.Cross(p21)#/d_p21_2
        return R


    def part_instant_rotation_global_axis_point(self, part, istep):
        w = self.part_global_rotation_vector(part, istep)
        w2 = w.Dot(w)
        if math.isclose(w2, 0, abs_tol=1e-8):
            return None
        step = self.interpolate_step(istep)
        frame = self.mechanism.part_global_frame(part, step)

        for point in [vm.O3D, 0.1*vm.X3D, 0.1*vm.Y3D, 0.1*vm.Z3D]:
            vp = self.part_local_point_global_speed(part, point, istep)

            if not math.isclose(vp.Norm(), 0, abs_tol=1e-6):
                return frame.OldCoordinates(point) - w.Cross(vp)/w2
        raise ValueError


    def plot2D(self, x=vm.X3D, y=vm.Y3D, isteps=None, plot_frames=False,
               plot_rotation_axis=False):
        fig, ax = plt.subplots()

        # Linkage colors
        np = len(self.mechanism.parts)
        colors = {p: hsv_to_rgb((ip / np, 0.78, 0.87)) for ip, p in enumerate(self.mechanism.parts)}
        colors[self.mechanism.ground] = (0,0,0)

#            i: to_hex(
#                ) for i in range(nlines)}

        if isteps == None:
            steps = self.steps[:]
        else:
            steps = [self.steps[i] for i in isteps]

#        # Fetching wireframes lines
#        wireframes = {}
#        for part in self.mechanism.parts:
#            # Fetching local points
#            part_points = []
#            for linkage in self.mechanism.part_linkages:


        for istep, step in enumerate(steps):
            linkage_positions = {}
            part_frames = {}
            for linkage in self.mechanism.linkages:

    #            flp1.origin.PlaneProjection2D(x, y).MPLPlot(ax=ax)
                if linkage.positions_require_kinematic_parameters:
                    ql = self.mechanism.extract_linkage_parameters_values(linkage,
                                                                          step)
                else:
                    ql = []

                part1_frame = self.mechanism.part_global_frame(linkage.part1,
                                                        step)
                part_frames[linkage.part1] = part1_frame
    #
                linkage_position1 = part1_frame.OldCoordinates(linkage.part1_position_function(ql))
                linkage_position1_2D = linkage_position1.PlaneProjection2D(x, y)

                part2_frame = self.mechanism.part_global_frame(linkage.part2,
                                                        step)
                part_frames[linkage.part2] = part2_frame
    #
                linkage_position2 = part2_frame.OldCoordinates(linkage.part2_position_function(ql))
                linkage_position2_2D = linkage_position1.PlaneProjection2D(x, y)

                if linkage_position1 != linkage_position2:
                    ax.text(*linkage_position1_2D, linkage.name+' position1')
                    ax.text(*linkage_position2_2D, linkage.name+' position2')
                    error = linkage_position2_2D - linkage_position1_2D
                    ax.add_patch(Arrow(*linkage_position1_2D,
                                       *error, 0.05))
                else:
                    if istep == 0:
                        ax.text(*linkage_position1_2D, linkage.name)

                linkage_positions[linkage, linkage.part1] = linkage_position1
                linkage_positions[linkage, linkage.part2] = linkage_position2


            part_linkages = self.mechanism.part_linkages()
            del part_linkages[self.mechanism.ground]

            for ipart, (part, linkages) in enumerate(part_linkages.items()):
#                middle_point = vm.o2D
#                for linkage in linkages:
#                    middle_point += linkage_positions[linkage, part]
#                for point in part.interest_points:
#                    middle_point += point
#                middle_point /= (len(linkages) + len(part.interest_points))
#                xm, ym = middle_point.vector
                points = []
                for linkage in linkages:
                    points.append(linkage_positions[linkage, part])
                points.extend([part_frames[part].OldCoordinates(p) for p in part.interest_points])
                xm, ym = vm.Point3D.mean_point(points).PlaneProjection2D(x, y).vector

                if istep == 0:
                    ax.text(xm, ym, part.name + ' step 0',
                            ha="center", va="center",
                            bbox=dict(boxstyle="square",
                                      ec=colors[part],
                                      fc=(1., 1, 1),
                                      ))
                else:
                    if ipart == 0:
                        ax.text(xm, ym, 'step {}'.format(istep),
                                ha="center", va="center",
                                bbox=dict(boxstyle="square",
                                      ec=colors[part],
                                      fc=(1., 1, 1),
                                      ))

#                for linkage in linkages:
#                    x1, y1 = linkage_positions[linkage, part]
#                    ax.plot([x1, xm], [y1, ym], color=colors[part])

                for line in Part.wireframe_lines(points):
                    line.MPLPlot2D(x, y, ax, color=colors[part], width=5)

                part_frame = self.mechanism.part_global_frame(part, step)
                for point in part.interest_points:
                    x1, y1 = part_frame.OldCoordinates(point).PlaneProjection2D(x, y)
                    ax.plot([x1, xm], [y1, ym], color=colors[part])


                if plot_frames:
                    part_frame = self.mechanism.part_global_frame(part, step)
                    part_frame.plot2d(x=x, y=y, ax=ax)

                if plot_rotation_axis:
                    axis = self.part_global_rotation_vector(part, istep)
                    point = self.part_instant_rotation_global_axis_point(part, istep)
                    if point is not None:
                        axis.Normalize()
                        line = vm.Line3D(point-axis, point+axis)
                        line.PlaneProjection2D(x, y).MPLPlot(ax=ax, color=colors[part], dashed=True)


        ax.set_aspect('equal')
        ax.set_xlabel(str(x))
        ax.set_ylabel(str(y))
        ax.margins(.1)

    def babylonjs(self, page='gm_babylonjs', plot_frames=False,
                  plot_trajectories=True, plot_instant_rotation_axis=False,
                  use_cdn=False):


        page+='.html'

        env = Environment(loader=PackageLoader('genmechanics', 'templates'),
                          autoescape=select_autoescape(['html', 'xml']))

        template = env.get_template('babylon.html')

        np = len(self.mechanism.parts)
        colors = {p: hsv_to_rgb((ip / np, 0.78, 0.87)) for ip, p in enumerate(self.mechanism.parts)}

        part_points = {p: [] for p in self.mechanism.parts}
        part_points[self.mechanism.ground] = []
#        part_frames = {}

        for part, linkages in self.mechanism.part_linkages().items():

            for linkage in linkages:
                if linkage.positions_require_kinematic_parameters:
                    ql = self.linkage_steps_parameters[0][linkage]
                else:
                    ql = []

                if part == linkage.part1:
                    linkage_position = linkage.part1_position_function(ql)
                else:
                    linkage_position = linkage.part2_position_function(ql)

                part_points[part].append(linkage_position)

            for point in part.interest_points:
                part_points[part].append(point)


        meshes_string = 'var parts_parent = [];\n'



        for part in self.mechanism.parts:
            meshes_string += 'var part_children = [];\n'
            lines = part.wireframe_lines(part_points[part])
            meshes_string += lines[0].babylon_script(name='part_parent', color=colors[part])
            meshes_string += 'parts_parent.push(part_parent);\n'
            for l in lines[1:]:
                meshes_string += l.Babylon(color=colors[part], parent='part_parent')
#                meshes_string += 'part_meshes.push(line);\n'

#            # Adding interest points
#            for point in part.interest_points:
#                meshes_string += 'var point = BABYLON.MeshBuilder.CreateSphere("interest_point", {diameter: 0.01}, scene);\n'
#                meshes_string += 'point.position = new BABYLON.Vector3({}, {}, {});'.format(*point.vector)
#                meshes_string += 'part_meshes.push(point);'


            if plot_frames:
                meshes_string += vm.OXYZ.babylonjs(parent='part_parent', size=0.1)

            # if plot_instant_rotation_axis:
            #     rotation_axis = self.par

        if plot_instant_rotation_axis:
            for part in self.mechanism.parts:
                line = vm.edges.LineSegment3D(-0.5*vm.X3D, 0.5*vm.X3D)
                meshes_string += line.babylon_script(name='rotation_axis',  color=colors[part], type_='dashed')
                meshes_string += 'parts_parent.push(rotation_axis);\n'


        linkages_string = ''
        for linkage in self.mechanism.linkages:
            if linkage not in self.mechanism.opened_linkages:

                ql = self.linkage_steps_parameters[0][linkage]
            else:
                ql = []

            if linkage.part1 in self.mechanism.parts:
                part1_parent = 'parts_parent[{}]'.format(self.mechanism.parts.index(linkage.part1))
            else:
                part1_parent = None

            if linkage.part2 in self.mechanism.parts:
                part2_parent = 'parts_parent[{}]'.format(self.mechanism.parts.index(linkage.part2))
            else:
                part2_parent = None

            linkages_string += linkage.babylonjs(ql,
                                                 part1_parent=part1_parent,
                                                 part2_parent=part2_parent)


        # Computing positions and orientations
        positions = []
        orientations = []
        linkage_positions = []

        # n_steps = len(self.steps)
        print(self.linkage_steps_parameters)
        for istep, step in enumerate(self.linkage_steps_parameters):
            step_positions = []
            step_orientations = []
            step_linkage_positions = []
#            step_linkage_positions = []
            for part in self.mechanism.parts:
                print(step)
                frame = round(self.mechanism.part_global_frame(part,
                                                  step))

                step_positions.append(list(frame.origin))
                step_orientations.append([list(frame.u),
                                          list(frame.v),
                                          list(frame.w)])


            if plot_instant_rotation_axis:
                # if istep == n_steps-1:
                #     istep_speed = n_steps-2
                #     # step_speed = self.steps[istep_speed]
                # else:
                #     istep_speed = istep
                #     # step_speed = step
                for part in self.mechanism.parts:

                    # print('unorm', u, u.Norm())
                    axis_point = self.part_instant_rotation_global_axis_point(part, istep)
                    if axis_point is None:
                        u = vm.X3D.copy()
                        v = vm.Y3D.copy()
                        w = vm.Z3D.copy()
                        axis_point = vm.Point3D((100, 100, 100))
                    else:
                        u = self.part_global_rotation_vector(part, istep)
                        u.Normalize()
                        v = u.RandomUnitNormalVector()
                        w = u.Cross(v)

                    step_positions.append(list(axis_point))
                    step_orientations.append([list(u),
                                              list(v),
                                              list(w)])
                    # print(step_positions)
                    # print(step_orientations)

            for linkage in self.mechanism.linkages:
                step_linkage_positions.append(list(self.mechanism.linkage_global_position(linkage, step)))

            positions.append(step_positions)
            orientations.append(step_orientations)
            linkage_positions.append(step_linkage_positions)




        trajectories = []
        if plot_trajectories:
            for trajectory in self.trajectories.values():
                trajectories.append([list(p) for p in trajectory])
#            for point, part in self.mechanism.parts:
#                for point in part.interest_points:
#                    trajectories.append([list(p.vector) for p in self.trajectory(point, part, self.mechanism.ground)])


        script = template.render(center=(0, 0, 0),
                                 length=2*0.5,
                                 meshes_string=meshes_string,
                                 linkages_string=linkages_string,
                                 positions=positions,
                                 orientations=orientations,
                                 linkage_positions=linkage_positions,
                                 trajectories=trajectories,
                                 use_cdn=use_cdn)

        with open(page,'w') as file:
            file.write(script)

        webbrowser.open('file://' + os.path.realpath(page))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 16 15:15:09 2017


"""

from numpy import array,zeros
import genmechanics.geometry as geometry

import math

class KnownLoad:
    def __init__(self,part,position,euler_angles,forces,torques,name=''):
        self.part=part
        self.position=position
        self.euler_angles=euler_angles
        self.forces=array(forces)
        self.torques=array(torques)
        self.name=name
        
        self.P=geometry.Euler2TransferMatrix(*self.euler_angles) 
        
class UnknownLoad:
    """
    :param force_directions: a list of directions for force (0,1,2)
    :param torque_directions: a list of directions for torque (0,1,2)
    """
    def __init__(self,part,position,euler_angles,static_matrix,
                 static_behavior_occurence_matrix,static_behavior_nonlinear_eq_indices,
                 static_behavior_linear_eq,static_behavior_nonlinear_eq,
                 static_require_kinematic,name=''):
        
        self.part=part
        self.position=position
        self.euler_angles=euler_angles

        self.static_matrix=static_matrix

        self.static_behavior_occurence_matrix=static_behavior_occurence_matrix
        
        self.static_behavior_nonlinear_eq_indices=static_behavior_nonlinear_eq_indices
        self.static_behavior_linear_eq=static_behavior_linear_eq
        self.static_behavior_nonlinear_eq=static_behavior_nonlinear_eq
        self.static_require_kinematic=static_require_kinematic

        self.name=name
#        print(euler_angles)
        self.P=geometry.Euler2TransferMatrix(*self.euler_angles) 
        self.n_static_unknowns=self.static_matrix.shape[1]


class SimpleUnknownLoad(UnknownLoad):
    """
    :param force_directions: a list of directions for force (0,1,2)
    :param torque_directions: a list of directions for torque (0,1,2)
    """
    def __init__(self,part,position,euler_angles,force_directions,torque_directions,name=''):
        
#        print(euler_angles)

        self.force_directions=force_directions
        self.torque_directions=torque_directions
        lfd=len(force_directions)
        ltd=len(torque_directions)
        static_matrix=zeros((6,lfd+ltd))
        for i,k in enumerate(force_directions):
            static_matrix[k,i]=1
        for i,k in enumerate(torque_directions):
            static_matrix[k+3,i+lfd]=1
            
        static_behavior_occurence_matrix=array([])
        static_behavior_occurence_matrix=array([])
        static_behavior_nonlinear_eq_indices=[]
        static_behavior_linear_eq=array([])
        static_behavior_nonlinear_eq=[]

        static_require_kinematic=False

        UnknownLoad.__init__(self,part,position,euler_angles,static_matrix,
                 static_behavior_occurence_matrix,static_behavior_nonlinear_eq_indices,
                 static_behavior_linear_eq,static_behavior_nonlinear_eq,
                 static_require_kinematic,name)
        

class GearSplashLoad(UnknownLoad):
    """
    Creates a splash force linked to the rotationnal speed around local X axis
    :param Rel: Reynolds limit
    :param d: caracteristic length. Default value: radius
    :param h: height of wet surface of gear
    """
    def __init__(self,part,position,euler_angles,h,radius,Cl,Ct,Rel,nu,d=None,name='Splash load'):
        self.Cl=Cl
        self.Ct=Ct
        self.Rel=Rel# limit reynods number
        self.radius=radius
        self.nu=nu
        
        if d==None:
            d=radius            
        self.d=d
            
        self.h=h

        if h<0:
            self.area=0
        elif h<self.radius:
            hp=self.radius-h
            self.area=self.radius**2*math.acos(hp/self.radius)-hp*math.sqrt(self.radius**2-hp**2)
        elif h<2*self.radius:
            h2=2*self.radius-h
            hp=self.radius-h2
            self.area=math.pi*self.radius**2-(self.radius**2*math.acos(hp/self.radius)-hp*math.sqrt(self.radius**2-hp**2))
        else:
            self.area=math.pi*self.radius**2
            
        
        self.nuSR=self.area*self.nu*self.radius
        self.wl=self.Rel*self.nu/self.radius/self.d
        
        static_matrix=zeros((6,1))
        static_matrix[3,0]=1# Resistant torque on X
        static_behavior_occurence_matrix=array([[1]])
        static_behavior_nonlinear_eq_indices=[0]
        static_behavior_linear_eq=array([])
        static_behavior_nonlinear_eq=[lambda x,w,v:
            x[0]+self.Cl*self.nuSR*w[0]
            if abs(w[0])<self.wl
            else x[0]+self.nuSR*(self.Ct*w[0]+(self.Cl-self.Ct)*self.wl*w[0]/abs(w[0]))]
        static_require_kinematic=True
        
        UnknownLoad.__init__(self,part,position,euler_angles,static_matrix,
                 static_behavior_occurence_matrix,static_behavior_nonlinear_eq_indices,
                 static_behavior_linear_eq,static_behavior_nonlinear_eq,
                 static_require_kinematic,name)        
        
        
    def ChangeCoefficients(self,Cl,Ct,Rel):
        self.Cl=Cl
        self.Ct=Ct
        self.Rel=Rel# limit reynods number
        self.wl=self.Rel*self.nu/self.radius/self.d
        self.static_behavior_nonlinear_eq=[lambda x,w,v:
            x[0]+self.Cl*self.nuSR*w[0]
            if abs(w[0])<self.wl
            else x[0]+self.nuSR*(self.Ct*w[0]+(self.Cl-self.Ct)*self.wl*w[0]/abs(w[0]))]
